""" Class for the LCG File Catalog Client

"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id"

from past.builtins import long
import six
from stat import *
import os
import re
import time

import DIRAC
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Resources.Catalog.Utilities import checkCatalogArguments
from DIRAC.Core.Utilities.Time import fromEpoch
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Core.Security.ProxyInfo import getProxyInfo, formatProxyInfoAsString
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import (
    getDNForUsername,
    getVOMSAttributeForGroup,
    getVOForGroup,
    getVOOption,
)
from DIRAC.Resources.Catalog.FileCatalogClientBase import FileCatalogClientBase

lfc = None
importedLFC = None

####################################################################
#
# These are some functions used by all methods in the class
#


def setLfnReplicas(lfn, replicas, successful, failed):
    if not lfn:
        return
    if replicas:
        successful[lfn] = replicas.copy()
        replicas.clear()
    elif lfn not in failed:
        failed[lfn] = "No active replica"


def getClientCertInfo():
    res = getProxyInfo(False, False)
    if not res["OK"]:
        gLogger.error("getClientCertInfo: Failed to get client proxy information.", res["Message"])
        return res
    proxyInfo = res["Value"]
    gLogger.debug(formatProxyInfoAsString(proxyInfo))
    if "group" not in proxyInfo:
        errStr = "getClientCertInfo: Proxy information does not contain the group."
        gLogger.error(errStr)
        return S_ERROR(errStr)
    if "VOMS" not in proxyInfo:
        proxyInfo["VOMS"] = getVOMSAttributeForGroup(proxyInfo["group"])
        errStr = "getClientCertInfo: Proxy information does not contain the VOMs information."
        gLogger.warn(errStr)
    res = getDNForUsername(proxyInfo["username"])
    if not res["OK"]:
        errStr = "getClientCertInfo: Error getting known proxies for user."
        gLogger.error(errStr, res["Message"])
        return S_ERROR(errStr)
    diracGroup = proxyInfo.get("group", "Unknown")
    vo = getVOForGroup(diracGroup)
    vomsVO = getVOOption(vo, "VOMSVO", "")
    resDict = {
        "DN": proxyInfo["identity"],
        "Role": proxyInfo["VOMS"],
        "User": proxyInfo["username"],
        "AllDNs": res["Value"],
        "Group": diracGroup,
        "VO": vo,
        "VOMSVO": vomsVO,
    }
    return S_OK(resDict)


def existsGuid(guid):
    """Check if the guid exists"""
    fstat = lfc.lfc_filestatg()
    error = lfc.lfc_statg("", guid, fstat)
    return returnCode(error and lfc.cvar.serrno != 2, not error)


def getDNFromUID(userID):
    buff = " " * (lfc.CA_MAXNAMELEN + 1)
    res = lfc.lfc_getusrbyuid(userID, buff)
    if res == 0:
        dn = buff[: buff.find("\x00")]
        gLogger.debug("LcgFileCatalogClient.getDNFromUID: UID %s maps to %s." % (userID, dn))
        return S_OK(dn)
    else:
        errStr = "LcgFileCatalogClient.getDNFromUID: Failed to get DN from UID"
        gLogger.error(errStr, "%s %s" % (userID, lfc.sstrerror(lfc.cvar.serrno)))
        return S_ERROR(errStr)


def getRoleFromGID(groupID, path=None):
    buff = " " * (lfc.CA_MAXNAMELEN + 1)
    res = lfc.lfc_getgrpbygid(groupID, buff)
    if res == 0:
        role = buff[: buff.find("\x00")]
        if role == "lhcb":
            role = "lhcb/Role=user"
        gLogger.debug("LcgFileCatalogClient.getRoleFromGID: GID %s maps to %s." % (groupID, role))
        return S_OK(role)
    else:
        errStr = "LcgFileCatalogClient:getRoleFromGID: Failed to get role from GID"
        gLogger.error(errStr, "%s %s%s" % (groupID, ("(%s) " % path) if path else "", lfc.sstrerror(lfc.cvar.serrno)))
        return S_ERROR()


def addReplica(guid, pfn, se, master):
    fid = lfc.lfc_fileid()
    status = "U"
    f_type = "D"
    poolname = ""
    fs = ""
    error = lfc.lfc_addreplica(guid, fid, se, pfn, status, f_type, poolname, fs)
    # If replica exists, re-register it as one may have changed some parameters (status, se)
    if lfc.sstrerror(lfc.cvar.serrno) == "File exists":
        retCode = removeReplica(pfn)
        if not retCode["OK"]:
            return retCode
        error = lfc.lfc_addreplica(guid, fid, se, pfn, status, f_type, poolname, fs)
    return returnCode(error)


def removeReplica(pfn):
    fid = lfc.lfc_fileid()
    error = lfc.lfc_delreplica("", fid, pfn)
    return returnCode(error and error != 2)


def setReplicaStatus(pfn, status):
    return returnCode(lfc.lfc_setrstatus(pfn, status))


def modReplica(pfn, newse):
    return returnCode(lfc.lfc_modreplica(pfn, "", "", newse))


def closeDirectory(oDirectory):
    return returnCode(lfc.lfc_closedir(oDirectory))


def getDNUserID(dn):
    error, users = lfc.lfc_getusrmap()
    userid = None
    for userMap in users if not error else []:
        if userMap.username == dn:
            userid = userMap.userid
            break
    return returnCode(userid is None, userid, errMsg="DN does not exist" if not error else "")


def addUserDN(userID, dn):
    error = lfc.lfc_enterusrmap(userID, dn)
    # 17 is if dn already exists, then OK
    return returnCode(error and lfc.cvar.serrno != 17)


def returnCode(error, value="", errMsg=""):
    if not error:
        return S_OK(value)
    elif errMsg:
        return S_ERROR(errMsg)
    else:
        return S_ERROR(lfc.sstrerror(lfc.cvar.serrno))


#####################################################
#
# LFC catalog client class
#
#####################################################


class LcgFileCatalogClient(FileCatalogClientBase):

    READ_METHODS = FileCatalogClientBase.READ_METHODS + [
        "isLink",
        "readLink",
        "isFile",
        "getFileMetadata",
        "getReplicas",
        "getReplicaStatus",
        "getFileSize",
        "isDirectory",
        "getDirectoryReplicas",
        "listDirectory",
        "getDirectoryMetadata",
        "getDirectorySize",
        "getDirectoryContents",
        "resolveDataset",
        "getLFNForPFN",
        "getUserDirectory",
    ]

    WRITE_METHODS = FileCatalogClientBase.WRITE_METHODS + [
        "createLink",
        "removeLink",
        "addFile",
        "addReplica",
        "removeReplica",
        "removeFile",
        "setReplicaStatus",
        "setReplicaHost",
        "createDirectory",
        "removeDirectory",
        "removeDataset",
        "removeFileFromDataset",
        "createDataset",
        "changePathOwner",
        "changePathMode",
    ]

    NO_LFN_METHODS = FileCatalogClientBase.NO_LFN_METHODS + [
        "getUserDirectory",
        "createUserDirectory",
        "createUserMapping",
        "removeUserDirectory",
    ]

    ADMIN_METHODS = FileCatalogClientBase.ADMIN_METHODS + [
        "getUserDirectory",
        "createUserDirectory",
        "createUserMapping",
        "removeUserDirectory",
    ]

    def __init__(self, **options):
        global lfc, importedLFC

        if importedLFC is None:
            try:
                import lfcthr as lfc

                # This is necessary to make the LFC client thread safe.
                lfc.init()
                importedLFC = True
                gLogger.debug("LcgFileCatalogClient.__init__: Successfully imported lfc module.")

            except ImportError:
                importedLFC = False
                gLogger.exception("LcgFileCatalogClient.__init__: Failed to import lfc module.")

        self.valid = importedLFC

        infosys = options.get("LcgGfalInfosys")
        if not infosys and "LCG_GFAL_INFOSYS" in os.environ:
            # if not in CS take from environ
            infosys = os.environ["LCG_GFAL_INFOSYS"]

        host = options.get("MasterHost")
        if not host and "LFC_HOST" in os.environ:
            # if not in CS take from environ
            host = os.environ["LFC_HOST"]

        self.host = host

        if host:
            os.environ["LFC_HOST"] = host
        if infosys:
            os.environ["LCG_GFAL_INFOSYS"] = infosys

        if "LFC_CONRETRYINT" not in os.environ:
            os.environ["LFC_CONRETRYINT"] = "5"
        if "LFC_CONNTIMEOUT" not in os.environ:
            os.environ["LFC_CONNTIMEOUT"] = "5"
        if "LFC_CONRETRY" not in os.environ:
            os.environ["LFC_CONRETRY"] = "5"

        self.prefix = "/grid"
        self.session = False
        self.transaction = False

    def isOK(self):
        return self.valid

    def getName(self):
        return S_OK(self.name)

    ####################################################################
    #
    # These are the methods for session/transaction manipulation
    #

    def __openSession(self):
        """Open the LFC client/server session"""
        if self.session:
            # Another thread is holding a session, can't create it
            return 0
        else:
            sessionName = "DIRAC_%s.%s at %s at time %s" % (
                DIRAC.majorVersion,
                DIRAC.minorVersion,
                DIRAC.siteName(),
                time.time(),
            )
            rc = lfc.lfc_startsess(self.host, sessionName)
            self.session = rc == 0
            # if there was an error, return -1, to be tested just after the call...
            return 1 if self.session else -1

    def __closeSession(self):
        """Close the LFC client/server session"""
        if self.session:
            lfc.lfc_endsess()
            self.session = False

    def __startTransaction(self):
        """Begin transaction for one time commit"""
        if not self.transaction:
            transactionName = "Transaction: DIRAC_%s.%s at %s at time %s" % (
                DIRAC.majorVersion,
                DIRAC.minorVersion,
                DIRAC.siteName(),
                time.time(),
            )
            lfc.lfc_starttrans(self.host, transactionName)
            self.transaction = True

    def __abortTransaction(self):
        """Abort transaction"""
        if self.transaction:
            lfc.lfc_aborttrans()
            self.transaction = False

    def __endTransaction(self):
        """End transaction gracefully"""
        if self.transaction:
            lfc.lfc_endtrans()
            self.transaction = False

    def setAuthorizationId(self, dn):
        """Set authorization id for the proxy-less LFC communication"""
        lfc.lfc_client_setAuthorizationId(0, 0, "GSI", dn)

    ####################################################################
    #
    # The following are read methods for paths
    #

    @checkCatalogArguments
    def exists(self, lfns):
        """Check if the path exists"""
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for lfn, guid in lfns.items():
            res = self.__existsLfn(lfn)
            if not res["OK"]:
                failed[lfn] = res["Message"]
            elif res["Value"]:
                successful[lfn] = lfn
            elif not isinstance(guid, six.string_types):
                successful[lfn] = False
            else:
                res = existsGuid(guid)
                if not res["OK"]:
                    failed[lfn] = res["Message"]
                elif not res["Value"]:
                    successful[lfn] = False
                else:
                    successful[lfn] = self.__getLfnForGUID(guid)["Value"]
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def __getPathAccess(self, path):
        """Determine the permissions using the lfc function lfc_access"""
        permDict = {"Read": 1, "Write": 2, "Execute": 4}
        resDict = {}
        for p in permDict:

            code = permDict[p]
            value = lfc.lfc_access(self.__fullLfn(path), code)
            if value == 0:
                resDict[p] = True
            else:
                resDict[p] = False
        return S_OK(resDict)

    @checkCatalogArguments
    def hasAccess(self, lfns, opType):

        if opType in LcgFileCatalogClient.READ_METHODS:
            opType = "Read"
        elif opType in LcgFileCatalogClient.WRITE_METHODS:
            opType = "Write"

        res = self.getPathPermissions(lfns)

        if not res["OK"]:
            return res

        perms = res["Value"]
        failed = perms["Failed"]
        successful = dict((path, perms["Successful"][path].get(opType, False)) for path in perms["Successful"])

        return S_OK({"Successful": successful, "Failed": failed})

    @checkCatalogArguments
    def getPathPermissions(self, lfns):
        """Determine the VOMs based ACL information for a supplied path"""
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for path in lfns:
            res = self.__getBasePath(path)
            if not res["OK"]:
                failed[path] = res["Message"]
            else:
                basePath = res["Value"]
                res = self.__getPathAccess(basePath)
                if not res["OK"]:
                    failed[path] = res["Message"]
                else:
                    lfcPerm = res["Value"]
                    res = self.__getACLInformation(basePath)
                    if not res["OK"]:
                        failed[path] = res["Message"]
                    else:
                        # Evaluate access rights
                        val = res["Value"]
                        try:
                            lfcPerm["user"] = val["user"]
                            lfcPerm["group"] = val["group"]
                            lfcPerm["world"] = val["world"]
                            lfcPerm["DN"] = val["DN"]
                            lfcPerm["Role"] = val["Role"]
                        except KeyError:
                            print("key not found: __getACLInformation returned incomplete dictionary", KeyError)
                            failed[path] = lfcPerm
                            continue
                    # ACLs are just an additional information, therefore here it is successful
                    successful[path] = lfcPerm

        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    ####################################################################
    #
    # The following are read methods for files
    #

    @checkCatalogArguments
    def isFile(self, lfns):

        # If we have less than three lfns to query a session doesn't make sense
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for lfn in lfns:
            res = self.__getPathStat(lfn)
            if not res["OK"]:
                failed[lfn] = res["Message"]
            elif S_ISREG(res["Value"].filemode):
                successful[lfn] = True
            else:
                successful[lfn] = False
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def getFileMetadata(self, lfns, ownership=False):
        """Returns the file metadata associated to a supplied LFN"""
        # If we have less than three lfns to query a session doesn't make sense
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for lfn in lfns:
            res = self.__getPathStat(lfn)
            if not res["OK"]:
                failed[lfn] = res["Message"]
            else:
                fstat = res["Value"]
                successful[lfn] = {}
                successful[lfn]["Size"] = fstat.filesize
                successful[lfn]["ChecksumType"] = fstat.csumtype
                successful[lfn]["Checksum"] = fstat.csumvalue
                successful[lfn]["GUID"] = fstat.guid
                successful[lfn]["Status"] = fstat.status
                successful[lfn]["CreationDate"] = fromEpoch(fstat.ctime)
                successful[lfn]["ModificationDate"] = fromEpoch(fstat.mtime)
                successful[lfn]["NumberOfLinks"] = fstat.nlink
                successful[lfn]["Mode"] = S_IMODE(fstat.filemode)
                if ownership:
                    res = getDNFromUID(fstat.uid)
                    if res["OK"]:
                        successful[lfn]["OwnerDN"] = res["Value"]
                    else:
                        successful[lfn]["OwnerDN"] = None
                    res = getRoleFromGID(fstat.gid, path=lfn)
                    if res["OK"]:
                        successful[lfn]["OwnerRole"] = res["Value"]
                    else:
                        successful[lfn]["OwnerRole"] = None
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def getFileSize(self, lfns):
        """Get the size of a supplied file"""
        # If we have less than three lfns to query a session doesn't make sense
        created = False
        if len(lfns) > 2:
            created = self.__openSession()
            if created < 0:
                return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for lfn in lfns:
            res = self.__getPathStat(lfn)
            if not res["OK"]:
                failed[lfn] = res["Message"]
            else:
                successful[lfn] = res["Value"].filesize
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def getReplicas(self, lfns, allStatus=False):
        """Returns replicas for an LFN or list of LFNs"""
        lfnChunks = breakListIntoChunks(lfns, 1000)
        # If we have less than three groups to query a session doesn't make sense
        created = False
        if len(lfnChunks) > 2:
            created = self.__openSession()
            if created < 0:
                return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for lfnList in lfnChunks:
            fullLfnList = [self.__fullLfn(lfn) for lfn in lfnList if lfn]
            value, replicaList = lfc.lfc_getreplicasl(fullLfnList, "")
            if value != 0:
                for lfn in lfnList:
                    reason = lfc.sstrerror(lfc.cvar.serrno)
                    if "Could not secure the connection" in reason:
                        # This is a fatal error
                        return S_ERROR("Could not secure the connection")
                    elif "Bad credentials" in reason:
                        return S_ERROR("Bad Credentials")
                continue
            guid = ""
            it = iter(lfnList)
            replicas = {}
            # This is useless but makes pylinit happy as lfn is defined in the loop when the guid changes
            lfn = None
            for oReplica in replicaList:
                if oReplica.errcode != 0:
                    if (oReplica.guid == "") or (oReplica.guid != guid):
                        setLfnReplicas(lfn, replicas, successful, failed)
                        lfn = it.next()
                        failed[lfn] = lfc.sstrerror(oReplica.errcode)
                        guid = oReplica.guid
                elif oReplica.sfn == "":
                    setLfnReplicas(lfn, replicas, successful, failed)
                    lfn = it.next()
                    failed[lfn] = "File has zero replicas"
                    guid = oReplica.guid
                else:
                    # This is where we change lfn for good!
                    if oReplica.guid != guid:
                        setLfnReplicas(lfn, replicas, successful, failed)
                        lfn = it.next()
                        guid = oReplica.guid
                    if (oReplica.status != "P") or allStatus:
                        se = oReplica.host
                        pfn = oReplica.sfn  # .strip()
                        replicas[se] = pfn
            # This is for the last file in the list
            setLfnReplicas(lfn, replicas, successful, failed)
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def getReplicaStatus(self, lfns):
        # If we have less than three lfns to query a session doesn't make sense
        created = False
        if len(lfns) > 2:
            created = self.__openSession()
            if created < 0:
                return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for lfn, se in lfns.items():
            res = self.__getFileReplicaStatus(lfn, se)
            if not res["OK"]:
                failed[lfn] = res["Message"]
            else:
                successful[lfn] = res["Value"]
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def getLFNForPFN(self, pfns):
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for pfn in pfns:
            res = self.__getLFNForPFN(pfn)
            if not res["OK"]:
                failed[pfn] = res["Message"]
            else:
                successful[pfn] = res["Value"]
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    ####################################################################
    #
    # The following a read methods for directories
    #

    @checkCatalogArguments
    def isDirectory(self, lfns):
        """Determine whether the path is a directory"""
        # If we have less than three lfns to query a session doesn't make sense
        created = False
        if len(lfns) > 2:
            created = self.__openSession()
            if created < 0:
                return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for lfn in lfns:
            res = self.__getPathStat(lfn)
            if not res["OK"]:
                failed[lfn] = res["Message"]
            elif S_ISDIR(res["Value"].filemode):
                successful[lfn] = True
            else:
                successful[lfn] = False
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def getDirectoryMetadata(self, lfns):
        # If we have less than three lfns to query a session doesn't make sense
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for lfn in lfns:
            res = self.__getPathStat(lfn)
            if not res["OK"]:
                failed[lfn] = res["Message"]
            else:
                fstat = res["Value"]
                successful[lfn] = {}
                successful[lfn]["Size"] = fstat.filesize
                successful[lfn]["ChecksumType"] = fstat.csumtype
                successful[lfn]["Checksum"] = fstat.csumvalue
                successful[lfn]["GUID"] = fstat.guid
                successful[lfn]["Status"] = fstat.status
                successful[lfn]["CreationDate"] = fromEpoch(fstat.ctime)
                successful[lfn]["ModificationDate"] = fromEpoch(fstat.mtime)
                successful[lfn]["NumberOfSubPaths"] = fstat.nlink
                res = getDNFromUID(fstat.uid)
                if res["OK"]:
                    successful[lfn]["OwnerDN"] = res["Value"]
                else:
                    successful[lfn]["OwnerDN"] = None
                res = getRoleFromGID(fstat.gid, path=lfn)
                if res["OK"]:
                    successful[lfn]["OwnerRole"] = res["Value"]
                else:
                    successful[lfn]["OwnerRole"] = None
                successful[lfn]["Mode"] = S_IMODE(fstat.filemode)
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def getDirectoryReplicas(self, lfns, allStatus=False):
        """This method gets all of the pfns in the directory"""
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for path in lfns:
            res = self.__getDirectoryContents(path)
            if not res["OK"]:
                failed[path] = res["Message"]
            else:
                pathReplicas = {}
                files = res["Value"]["Files"]
                for lfn, fileDict in files.items():
                    pathReplicas[lfn] = {}
                    for se, seDict in fileDict["Replicas"].items():
                        pfn = seDict["PFN"]
                        status = seDict["Status"]
                        if (status != "P") or allStatus:
                            pathReplicas[lfn][se] = pfn
                successful[path] = pathReplicas
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def listDirectory(self, lfns, verbose=False):
        """Returns the result of __getDirectoryContents for multiple supplied paths"""
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for path in lfns:
            res = self.__getDirectoryContents(path, verbose)
            if res["OK"]:
                successful[path] = res["Value"]
            else:
                failed[path] = res["Message"]
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def getDirectorySize(self, lfns, longOutput=False, rawFiles=False):
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for path in lfns.keys():
            res = self.__getDirectorySize(path, longOutput=longOutput)
            if res["OK"]:
                successful[path] = res["Value"]
            else:
                failed[path] = res["Message"]
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    ####################################################################
    #
    # The following are read methods for links
    #

    @checkCatalogArguments
    def isLink(self, links):
        # If we have less than three lfns to query a session doesn't make sense
        failed = {}
        successful = {}
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        for link in links:
            res = self.__getLinkStat(link)
            if not res["OK"]:
                failed[link] = res["Message"]
            elif S_ISLNK(res["Value"].filemode):
                successful[link] = True
            else:
                successful[link] = False
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def readLink(self, links):
        # If we have less than three lfns to query a session doesn't make sense
        created = False
        if len(links) > 2:
            created = self.__openSession()
            if created < 0:
                return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for link in links:
            res = self.__readLink(link)
            if res["OK"]:
                successful[link] = res["Value"]
            else:
                failed[link] = res["Message"]
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    ####################################################################
    #
    # The following are read methods for datasets
    #

    @checkCatalogArguments
    def resolveDataset(self, datasets, allStatus=False):
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        successful = {}
        failed = {}
        for datasetName in datasets:
            res = self.__getDirectoryContents(datasetName)
            if not res["OK"]:
                failed[datasetName] = res["Message"]
            else:
                # linkDict = res['Value']['Links']
                linkDict = res["Value"]["Files"]
                datasetFiles = {}
                for link, fileMetadata in linkDict.items():
                    # target = fileMetadata[link]['MetaData']['Target']
                    target = link
                    datasetFiles[target] = fileMetadata["Replicas"]
                successful[datasetName] = datasetFiles
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    ####################################################################
    #
    # The following a write methods for files
    #

    @checkCatalogArguments
    def addFile(self, lfns):
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        baseDirs = []
        for lfn in lfns:
            baseDir = os.path.dirname(lfn)
            if baseDir in baseDirs:
                continue
            baseDirs.append(baseDir)
            res = self.__executeOperation(baseDir, "exists")
            # If we failed to find out whether the directory exists
            if not res["OK"]:
                continue
            # If the directory exists
            if res["Value"]:
                continue
            # Make the directories recursively if needed
            res = self.__makeDirs(baseDir)
            # If we failed to make the directory for the file
            if not res["OK"]:
                continue
        lfc.lfc_umask(0000)
        for lfnList in breakListIntoChunks(lfns, 1000):
            fileChunk = []
            for lfn in list(lfnList):
                lfnInfo = lfns[lfn]
                pfn = lfnInfo["PFN"]
                size = lfnInfo["Size"]
                se = lfnInfo["SE"]
                guid = lfnInfo["GUID"]
                checksum = lfnInfo["Checksum"]
                res = self.__checkAddFile(lfn, pfn, size, se, guid, checksum)
                if not res["OK"]:
                    # Error
                    failed[lfn] = res["Message"]
                    lfnList.remove(lfn)
                elif not res["Value"]:
                    # File already exists adn is consistent
                    successful[lfn] = True
                    lfnList.remove(lfn)
                else:
                    # File doesn't exist, create it
                    oFile = lfc.lfc_filereg()
                    oFile.lfn = self.__fullLfn(lfn)
                    oFile.sfn = pfn
                    oFile.size = size
                    oFile.mode = 0o664
                    oFile.server = se
                    oFile.guid = guid
                    oFile.csumtype = "AD"
                    oFile.status = "U"
                    oFile.csumvalue = lfnInfo["Checksum"]
                    fileChunk.append(oFile)
            if not lfnList:
                continue
            error, errCodes = lfc.lfc_registerfiles(fileChunk)
            if error or (len(errCodes) != len(lfnList)):
                for lfn in lfnList:
                    failed[lfn] = lfc.sstrerror(lfc.cvar.serrno)
                continue
            for index in range(len(errCodes)):
                lfn = lfnList[index]
                errCode = errCodes[index]
                if errCode == 0:
                    successful[lfn] = True
                elif errCode == 17:
                    failed[lfn] = "The supplied GUID is already used"
                    res = self.__getLfnForGUID(guid)
                    if res["OK"]:
                        failed[lfn] = "The supplied GUID is already used by %s" % res["Value"]
                else:
                    failed[lfn] = lfc.sstrerror(errCode)
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def addReplica(self, lfns):
        """This adds a replica to the catalogue."""
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for lfn, info in lfns.items():
            pfn = info["PFN"]
            se = info["SE"]
            if "Master" not in info:
                master = False
            else:
                master = info["Master"]
            res = self.__getLFNGuid(lfn)
            if not res["OK"]:
                failed[lfn] = res["Message"]
            else:
                guid = res["Value"]
                res = addReplica(guid, pfn, se, master)
                if res["OK"]:
                    successful[lfn] = True
                else:
                    failed[lfn] = res["Message"]
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def removeFile(self, lfns):
        """Remove the supplied path"""
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        res = self.exists(lfns)
        if not res["OK"]:
            if created:
                self.__closeSession()
            return res
        failed = res["Value"]["Failed"]
        successful = {}
        for lfn, exists in res["Value"]["Successful"].items():
            if not exists:
                successful[lfn] = True
            else:
                res = self.__unlinkPath(lfn)
                if res["OK"]:
                    successful[lfn] = True
                else:
                    failed[lfn] = res["Message"]
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def removeReplica(self, lfns):
        created = False
        if len(lfns) > 2:
            created = self.__openSession()
            if created < 0:
                return S_ERROR("Error opening LFC session")
        res = self.getReplicas(lfns)  # We need the PFNs of the input lfn (list)
        if not res["OK"]:
            return res
        for lfn, lfnrep in res["Value"]["Successful"].items():
            if "PFN" not in lfns[lfn]:  # Update only if the PFN was not supplied
                lfns[lfn]["PFN"] = lfnrep[lfns[lfn]["SE"]]
        failed = {}
        for lfn, message in res["Value"]["Failed"].items():
            if "PFN" not in lfns[lfn]:  # Change only if PFN is not there
                failed[lfn] = message  # The replicas are not available, mark the lfn as failed
                lfns.pop(lfn)  # and remove them
        successful = {}
        for lfn, info in lfns.items():
            if ("PFN" not in info) or ("SE" not in info):
                failed[lfn] = "Required parameters not supplied"
            else:
                pfn = info["PFN"]
                se = info["SE"]
                res = removeReplica(pfn)
                if res["OK"]:
                    successful[lfn] = True
                else:
                    if res["Message"] == "No such file or directory":
                        # The PFN didn't exist, but maybe it wsa changed...
                        res1 = self.getReplicas(lfn)
                        if res1["OK"]:
                            pfn1 = res1["Value"]["Successful"].get(lfn, {}).get(se)
                            if pfn1 and pfn1 != pfn:
                                res = removeReplica(pfn1)
                    if res["OK"]:
                        successful[lfn] = True
                    else:
                        failed[lfn] = res["Message"]
        lfnRemoved = list(successful)
        if len(lfnRemoved) > 0:
            res = self.getReplicas(lfnRemoved, True)
            zeroReplicaFiles = []
            if not res["OK"]:
                if created:
                    self.__closeSession()
                return res
            else:
                for lfn, repDict in res["Value"]["Successful"].items():
                    if len(repDict) == 0:
                        zeroReplicaFiles.append(lfn)
            if len(zeroReplicaFiles) > 0:
                res = self.removeFile(zeroReplicaFiles)
                if not res["OK"]:
                    if created:
                        self.__closeSession()
                    return res
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def setReplicaStatus(self, lfns):
        created = False
        if len(lfns) > 2:
            created = self.__openSession()
            if created < 0:
                return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for lfn, info in lfns.items():
            pfn = info["PFN"]
            status = info["Status"]
            res = setReplicaStatus(pfn, status[0])
            if res["OK"]:
                successful[lfn] = True
            else:
                failed[lfn] = res["Message"]
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def setReplicaProblematic(self, lfns, revert=False):
        """
        Set replicas to problematic.

        :param lfn lfns: has to be formated this way :
                    { lfn : { se1 : pfn1, se2 : pfn2, ...}, ...}
        :param revert: If True, remove the problematic flag

        :return: { Successful : { lfn : [ ses ] }, Failed : { lfn : { se : msg } } }
        """

        # This method does a batch treatment because the setReplicaStatus can only take one replica per lfn at once
        #
        # Illustration :
        #
        # lfns {'L2': {'S1': 'P3'}, 'L3': {'S3': 'P5', 'S2': 'P4', 'S4': 'P6'}, 'L1': {'S2': 'P2', 'S1': 'P1'}}
        #
        # loop1: lfnSEs {'L2': ['S1'], 'L3': ['S3', 'S2', 'S4'], 'L1': ['S2', 'S1']}
        # loop1 : batch {'L2': {'Status': 'P', 'SE': 'S1', 'PFN': 'P3'}, 'L3': {'Status': 'P', 'SE': 'S4', 'PFN': 'P6'},
        #                'L1': {'Status': 'P', 'SE': 'S1', 'PFN': 'P1'}}
        #
        # loop2: lfnSEs {'L2': [], 'L3': ['S3', 'S2'], 'L1': ['S2']}
        # loop2 : batch {'L3': {'Status': 'P', 'SE': 'S2', 'PFN': 'P4'}, 'L1': {'Status': 'P', 'SE': 'S2', 'PFN': 'P2'}}
        #
        # loop3: lfnSEs {'L3': ['S3'], 'L1': []}
        # loop3 : batch {'L3': {'Status': 'P', 'SE': 'S3', 'PFN': 'P5'}}
        #
        # loop4: lfnSEs {'L3': []}
        # loop4 : batch {}

        successful = {}
        failed = {}

        status = "-" if revert else "P"

        # { lfn : [ se1, se2, ...], ...}
        lfnsSEs = dict((lfn, [se for se in lfns[lfn]]) for lfn in lfns)

        while lfnsSEs:

            # { lfn : { 'SE' : se1, 'PFN' : pfn1, 'Status' : status }, ... }
            batch = {}

            for lfn in list(lfnsSEs):
                # If there are still some Replicas (SE) for the given LFN, we put it in the next batch
                # else we remove the entry from the lfnsSEs dict
                if lfnsSEs[lfn]:
                    se = lfnsSEs[lfn].pop()
                    batch[lfn] = {"SE": se, "PFN": lfns[lfn][se], "Status": status}
                else:
                    del lfnsSEs[lfn]

            # Happens when there is nothing to treat anymore
            if not batch:
                break

            res = self.setReplicaStatus(batch)
            if not res["OK"]:
                for lfn in batch:
                    failed.setdefault(lfn, {})[batch[lfn]["SE"]] = res["Message"]
                continue

            for lfn in res["Value"]["Failed"]:
                failed.setdefault(lfn, {})[batch[lfn]["SE"]] = res["Value"]["Failed"][lfn]

            for lfn in res["Value"]["Successful"]:
                successful.setdefault(lfn, []).append(batch[lfn]["SE"])

        return S_OK({"Successful": successful, "Failed": failed})

    @checkCatalogArguments
    def setReplicaHost(self, lfns):
        """This modifies the replica metadata for the SE."""
        created = False
        if len(lfns) > 2:
            created = self.__openSession()
            if created < 0:
                return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for lfn, info in lfns.items():
            pfn = info["PFN"]
            newse = info["NewSE"]
            res = modReplica(pfn, newse)
            if res["OK"]:
                successful[lfn] = True
            else:
                failed[lfn] = res["Message"]
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    ####################################################################
    #
    # The following are write methods for directories
    #

    @checkCatalogArguments
    def removeDirectory(self, lfns, recursive=False):
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        res = self.exists(lfns)
        if not res["OK"]:
            if created:
                self.__closeSession()
            return res
        failed = res["Value"]["Failed"]
        successful = {}
        for lfn, exists in res["Value"]["Successful"].items():
            if not exists:
                successful[lfn] = True
                continue
            if recursive:
                res = self.__removeDirs(lfn)
            else:
                res = self.__removeDirectory(lfn)
            if res["OK"]:
                successful[lfn] = True
            else:
                failed[lfn] = res["Message"]
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def createDirectory(self, lfns):
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for path in lfns:
            res = self.__makeDirs(path)
            if res["OK"]:
                successful[path] = True
            else:
                failed[path] = res["Message"]
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    ####################################################################
    #
    # The following are write methods for links
    #

    @checkCatalogArguments
    def createLink(self, links):
        # If we have less than three lfns to query a session doesn't make sense
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for link, target in links.items():
            res = self.__makeDirs(os.path.dirname(link))
            if not res["OK"]:
                failed[link] = res["Message"]
            else:
                res = self.__makeLink(link, target)
                if not res["OK"]:
                    failed[link] = res["Message"]
                else:
                    successful[link] = target
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def removeLink(self, links):
        # If we have less than three lfns to query a session doesn't make sense
        created = False
        if len(links) > 2:
            created = self.__openSession()
            if created < 0:
                return S_ERROR("Error opening LFC session")
        failed = {}
        successful = {}
        for link in links:
            res = self.__unlinkPath(link)
            if not res["OK"]:
                failed[link] = res["Message"]
            else:
                successful[link] = True
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    ####################################################################
    #
    # The following are write methods for datasets
    #

    @checkCatalogArguments
    def createDataset(self, datasets):
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        successful = {}
        failed = {}
        for datasetName, lfns in datasets.items():
            res = self.__executeOperation(datasetName, "exists")
            if not res["OK"]:
                return res
            elif res["Value"]:
                return S_ERROR("LcgFileCatalogClient.createDataset: This dataset already exists.")
            res = self.__createDataset(datasetName, lfns)
            if res["OK"]:
                successful[datasetName] = True
            else:
                self.__executeOperation(datasetName, "removeDataset")
                failed[datasetName] = res["Message"]
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def removeDataset(self, datasets):
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        successful = {}
        failed = {}
        for datasetName in datasets:
            res = self.__removeDataset(datasetName)
            if not res["OK"]:
                failed[datasetName] = res["Message"]
            else:
                successful[datasetName] = True
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def removeFileFromDataset(self, datasets):
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        successful = {}
        failed = {}
        for datasetName, lfns in datasets.items():
            res = self.__removeFilesFromDataset(datasetName, lfns)
            if not res["OK"]:
                failed[datasetName] = res["Message"]
            else:
                successful[datasetName] = True
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    ####################################################################
    #
    # These are the internal methods to be used by all methods
    #

    def __executeOperation(self, path, method):
        """Executes the requested functionality with the supplied path"""
        fcn = None
        if hasattr(self, method) and callable(getattr(self, method)):
            fcn = getattr(self, method)
        if not fcn:
            return S_ERROR("Unable to invoke %s, it isn't a member function of LcgFileCatalogClient" % method)
        res = fcn(path)
        if isinstance(path, dict):
            path = list(path)[0]
        if not res["OK"]:
            return res
        elif path not in res["Value"]["Successful"]:
            return S_ERROR(res["Value"]["Failed"][path])
        return S_OK(res["Value"]["Successful"][path])

    def __getLFNForPFN(self, pfn):
        fstat = lfc.lfc_filestatg()
        error = lfc.lfc_statr(pfn, fstat)
        return returnCode(error, self.__getLfnForGUID(fstat.guid) if not error else None)

    def __existsLfn(self, lfn):
        """Check whether the supplied LFN exists"""
        error = lfc.lfc_access(self.__fullLfn(lfn), 0)
        return returnCode(error and lfc.cvar.serrno != 2, error == 0)

    @checkCatalogArguments
    def getLFNForGUID(self, guids):
        guidLFN = {}
        failed = {}

        for guid in guids:
            # I somehow have the feeling that lfnlist[0] of __getLfnForGUID
            # could throw an exception if the guid does not exist in the DB
            # but... not touching this black magic
            try:
                guidLFN[guid] = self.__getLfnForGUID(guid)["Value"]
            except Exception as _e:
                failed[guid] = "GUID does not exist"
        return S_OK({"Successful": guidLFN, "Failed": failed})

    def __getLfnForGUID(self, guid):
        """Resolve the LFN for a supplied GUID"""
        if not guid:
            return S_OK()
        linkList = lfc.lfc_list()
        lfnlist = []
        listlinks = lfc.lfc_listlinks("", guid, lfc.CNS_LIST_BEGIN, linkList)
        while listlinks:
            ll = listlinks.path
            if re.search("^" + self.prefix, ll):
                ll = listlinks.path.replace(self.prefix, "", 1)
            lfnlist.append(ll)
            listlinks = lfc.lfc_listlinks("", guid, lfc.CNS_LIST_CONTINUE, linkList)
        else:
            lfc.lfc_listlinks("", guid, lfc.CNS_LIST_END, linkList)
        return S_OK(lfnlist[0])

    def __getBasePath(self, path):
        exists = False
        while not exists:
            res = self.__executeOperation(path, "exists")
            if not res["OK"]:
                return res
            else:
                exists = res["Value"]
                if not exists:
                    path = os.path.dirname(path)
        return S_OK(path)

    def __getACLInformation(self, path):
        results, objects = lfc.lfc_getacl(self.__fullLfn(path), 256)  # lfc.CNS_ACL_GROUP_OBJ)
        if results == -1:
            errStr = "LcgFileCatalogClient.__getACLInformation: Failed to obtain all path ACLs."
            gLogger.error(errStr, "%s %s" % (path, lfc.sstrerror(lfc.cvar.serrno)))
            return S_ERROR(errStr)
        permissionsDict = {}
        for obj in objects:
            if obj.a_type == lfc.CNS_ACL_USER_OBJ:
                res = getDNFromUID(obj.a_id)
                if not res["OK"]:
                    return res
                permissionsDict["DN"] = res["Value"]
                permissionsDict["user"] = obj.a_perm
            elif obj.a_type == lfc.CNS_ACL_GROUP_OBJ:
                res = getRoleFromGID(obj.a_id, path=path)
                if not res["OK"]:
                    return res
                role = res["Value"]
                permissionsDict["Role"] = role
                permissionsDict["group"] = obj.a_perm
            elif obj.a_type == lfc.CNS_ACL_OTHER:
                permissionsDict["world"] = obj.a_perm
            else:
                errStr = "LcgFileCatalogClient.__getACLInformation: ACL type not considered."
                gLogger.debug(errStr, obj.a_type)
        gLogger.verbose(
            "LcgFileCatalogClient.__getACLInformation: %s owned by %s:%s."
            % (path, permissionsDict["DN"], permissionsDict["Role"])
        )
        return S_OK(permissionsDict)

    def __getPathStat(self, path="", guid=""):
        if path:
            path = self.__fullLfn(path)
        fstat = lfc.lfc_filestatg()
        error = lfc.lfc_statg(path, guid, fstat)
        return returnCode(error, fstat)

    def __getFileReplicas(self, lfn, allStatus):
        error, replicaObjects = lfc.lfc_getreplica(self.__fullLfn(lfn), "", "")
        return returnCode(
            error or not replicaObjects,
            dict([(replica.host, replica.sfn) for replica in replicaObjects if allStatus or replica.status != "P"])
            if not error
            else None,
            errMsg="File has zero replicas" if not error else "",
        )

    def __getFileReplicaStatus(self, lfn, se):
        error, replicaObjects = lfc.lfc_getreplica(self.__fullLfn(lfn), "", "")
        status = None
        for replica in replicaObjects if not error else []:
            if se == replica.host:
                status = replica.status
                break
        return returnCode(status is None, status, errMsg="No replica at supplied site" if not error else "")

    def __checkAddFile(self, lfn, pfn, size, se, guid, checksum):
        res = self.__getPathStat(lfn)
        if not res["OK"]:
            if res["Message"] != "No such file or directory":
                return S_ERROR("Failed to find pre-existance of LFN")
        else:
            # File exists, check if consistent with supplied parameters
            fstat = res["Value"]
            errStr = ""
            if fstat.guid != guid:
                errStr = "This LFN %s is already registered with another GUID" % lfn
            elif fstat.filesize != size:
                errStr = "This LFN %s is already registered with another size" % lfn
            elif fstat.csumvalue.upper() != checksum.upper():
                errStr = "This LFN %s is already registered with another adler32" % lfn
            if errStr:
                return S_ERROR(errStr)
            res = self.__getFileReplicas(lfn, True)
            if not res["OK"]:
                return S_ERROR("Failed to obtain replicas for existing LFN %s" % lfn)
            replicas = res["Value"]
            if replicas.get(se) != pfn:
                return S_ERROR("This LFN %s is already registered with another SE/PFN" % lfn)
            return S_OK(False)
        # We reach here only if the file doesn't exist, which is what we look for!!
        # Now we check the arguments
        try:
            errStr = ""
            size = long(size)
        except Exception:
            errStr = "The size of the file must be an 'int','long' or 'string'"
        if not se:
            errStr = "The SE for the file was not supplied."
        elif not pfn:
            errStr = "The PFN for the file was not supplied."
        elif not lfn:
            errStr = "The LFN for the file was not supplied."
        elif not guid:
            errStr = "The GUID for the file was not supplied."
        elif not checksum:
            errStr = "The adler32 for the file was not supplied."
        if errStr:
            return S_ERROR(errStr)
        return S_OK(True)

    def __unlinkPath(self, lfn):
        return returnCode(lfc.lfc_unlink(self.__fullLfn(lfn)))

    def __removeDirectory(self, path):
        return returnCode(lfc.lfc_rmdir(self.__fullLfn(path)))

    def __removeDirs(self, path):
        """Black magic contained within..."""
        res = self.__getDirectoryContents(path)
        if not res["OK"]:
            return res
        subDirs = res["Value"]["SubDirs"]
        files = res["Value"]["Files"]
        for subDir in subDirs:
            res = self.__removeDirs(subDir)
            if not res["OK"]:
                return res
        if files:
            return S_ERROR("Directory not empty")
        return self.__removeDirectory(path)

    def __makeDirs(self, path, mode=0o775):
        """Black magic contained within...."""
        dirName = os.path.dirname(path)
        res = self.__executeOperation(path, "exists")
        if not res["OK"]:
            return res
        if res["Value"]:
            return S_OK()
        res = self.__executeOperation(dirName, "exists")
        if not res["OK"]:
            return res
        if res["Value"]:
            res = self.__makeDirectory(path, mode)
        else:
            res = self.__makeDirs(dirName, mode)
            res = self.__makeDirectory(path, mode)
        return res

    def __makeDirectory(self, path, mode):
        lfc.lfc_umask(0000)
        return returnCode(lfc.lfc_mkdir(self.__fullLfn(path), mode))

    def __openDirectory(self, path):
        value = lfc.lfc_opendirg(self.__fullLfn(path), "")
        return returnCode(not value, value)

    def __getDirectoryContents(self, path, verbose=False):
        """Returns a dictionary containing all of the contents of a directory.
        This includes the metadata associated to files (replicas, size, guid, status) and the subdirectories found.
        """
        # First check that the directory exists
        res = self.__executeOperation(path, "exists")
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR("No such file or directory")

        res = self.__getPathStat(path)
        if not res["OK"]:
            return res
        nbfiles = res["Value"].nlink
        res = self.__openDirectory(path)
        if not res["OK"]:
            return res
        oDirectory = res["Value"]
        subDirs = {}
        links = {}
        files = {}
        loop = list(range(nbfiles + 1))
        while loop.pop():
            result = lfc.lfc_readdirxr(oDirectory, "")
            if not result:
                # In some rare cases we reach the end of oDirectory, before nbfiles iterations (!!!)
                break
            entry, fileInfo = result
            pathMetadata = {}
            pathMetadata["Mode"] = S_IMODE(entry.filemode)
            subPath = "%s/%s" % (path, entry.d_name)
            if verbose:
                statRes = self.__getPathStat(subPath)
                if statRes["OK"]:
                    oPath = statRes["Value"]
                    pathMetadata["Size"] = oPath.filesize
                    pathMetadata["ChecksumType"] = oPath.csumtype
                    pathMetadata["Checksum"] = oPath.csumvalue
                    pathMetadata["GUID"] = oPath.guid
                    pathMetadata["Status"] = oPath.status
                    pathMetadata["CreationDate"] = fromEpoch(oPath.ctime)
                    pathMetadata["ModificationDate"] = fromEpoch(oPath.mtime)
                    pathMetadata["NumberOfLinks"] = oPath.nlink
                    pathMetadata["LastAccess"] = oPath.atime
                    res = getDNFromUID(oPath.uid)
                    if res["OK"]:
                        pathMetadata["OwnerDN"] = res["Value"]
                    else:
                        pathMetadata["OwnerDN"] = None
                    res = getRoleFromGID(oPath.gid, path=subPath)
                    if res["OK"]:
                        pathMetadata["OwnerRole"] = res["Value"]
                    else:
                        pathMetadata["OwnerRole"] = None
            if S_ISDIR(entry.filemode):
                subDirs[subPath] = pathMetadata
            else:
                replicaDict = {}
                if fileInfo:
                    for replica in fileInfo:
                        replicaDict[replica.host] = {"PFN": replica.sfn, "Status": replica.status}
                pathMetadata["Size"] = entry.filesize
                pathMetadata["GUID"] = entry.guid
                if S_ISLNK(entry.filemode):
                    res = self.__executeOperation(subPath, "readLink")
                    if res["OK"]:
                        pathMetadata["Target"] = res["Value"]
                    links[subPath] = {}
                    links[subPath]["MetaData"] = pathMetadata
                    links[subPath]["Replicas"] = replicaDict
                elif S_ISREG(entry.filemode):
                    files[subPath] = {}
                    files[subPath]["Replicas"] = replicaDict
                    files[subPath]["MetaData"] = pathMetadata
        pathDict = {}
        res = closeDirectory(oDirectory)
        pathDict = {"Files": files, "SubDirs": subDirs, "Links": links}
        return S_OK(pathDict)

    def __getDirectorySize(self, path, longOutput=False):
        res = self.__executeOperation(path, "exists")
        if not res["OK"]:
            return res
        if not res["Value"]:
            return S_ERROR("No such file or directory")

        res = self.__getPathStat(path)
        if not res["OK"]:
            return res
        nbfiles = res["Value"].nlink
        res = self.__openDirectory(path)
        if not res["OK"]:
            return res
        oDirectory = res["Value"]
        pathDict = {"SubDirs": {}, "ClosedDirs": [], "Files": 0, "TotalSize": 0, "SiteUsage": {}}
        loop = list(range(nbfiles + 1))
        while loop.pop():
            entry, fileInfo = lfc.lfc_readdirxr(oDirectory, "")
            if S_ISDIR(entry.filemode):
                subDir = "%s/%s" % (path, entry.d_name)
                permissions = S_IMODE(entry.filemode)
                if (not permissions & S_IWUSR) and (not permissions & S_IWGRP) and (not permissions & S_IWOTH):
                    pathDict["ClosedDirs"].append(subDir)
                modTime = time.ctime()
                statRes = self.__getPathStat(subDir)
                if statRes["OK"]:
                    modTime = fromEpoch(statRes["Value"].mtime)
                pathDict["SubDirs"][subDir] = modTime
            else:
                fileSize = entry.filesize
                pathDict["TotalSize"] += fileSize
                pathDict["Files"] += 1
                if not fileInfo:
                    gLogger.error(
                        "LcgFileCatalogClient.__getDirectorySize: File found with no replicas",
                        "%s/%s" % (path, entry.d_name),
                    )
                else:
                    for replica in fileInfo:
                        if replica.host not in pathDict["SiteUsage"]:
                            pathDict["SiteUsage"][replica.host] = {"Files": 0, "Size": 0}
                        pathDict["SiteUsage"][replica.host]["Size"] += fileSize
                        pathDict["SiteUsage"][replica.host]["Files"] += 1
        res = closeDirectory(oDirectory)
        return S_OK(pathDict)

    def __getLinkStat(self, link):
        lstat = lfc.lfc_filestat()
        return returnCode(lfc.lfc_lstat(self.__fullLfn(link), lstat), lstat)

    def __readLink(self, link):
        buff = " " * (lfc.CA_MAXPATHLEN + 1)
        chars = lfc.lfc_readlink(self.__fullLfn(link), buff, lfc.CA_MAXPATHLEN)
        if chars > 0:
            error = 0
            chars = buff[:chars].replace(self.prefix, "", 1).replace("\x00", "")
        else:
            error = 1
        return returnCode(error, chars)

    def __makeLink(self, source, target):
        return returnCode(lfc.lfc_symlink(self.__fullLfn(target), self.__fullLfn(source)))

    def __getLFNGuid(self, lfn):
        """Get the GUID for the given lfn"""
        fstat = lfc.lfc_filestatg()
        return returnCode(lfc.lfc_statg(self.__fullLfn(lfn), "", fstat), fstat.guid)

    def __createDataset(self, datasetName, lfns):
        res = self.__makeDirs(datasetName)
        if not res["OK"]:
            return res
        links = {}
        for lfn in lfns:
            res = self.__getLFNGuid(lfn)
            if not res["OK"]:
                return res
            else:
                link = "%s/%s" % (datasetName, res["Value"])
                links[link] = lfn
        res = self.createLink(links)
        if len(res["Value"]["Successful"]) == len(links):
            return S_OK()
        totalError = ""
        for link, error in res["Value"]["Failed"].items():
            gLogger.error("LcgFileCatalogClient.__createDataset: Failed to create link", "for %s: %s" % (link, error))
            totalError = "%s\n %s : %s" % (totalError, link, error)
        return S_ERROR(totalError)

    def __removeDataset(self, datasetName):
        res = self.__getDirectoryContents(datasetName)
        if not res["OK"]:
            return res
        links = list(res["Value"]["Files"])
        res = self.removeLink(links)
        if not res["OK"]:
            return res
        elif len(res["Value"]["Failed"]):
            return S_ERROR("Failed to remove all links")
        else:
            res = self.__executeOperation(datasetName, "removeDirectory")
            return res

    def __removeFilesFromDataset(self, datasetName, lfns):
        links = []
        for lfn in lfns:
            res = self.__getLFNGuid(lfn)
            if not res["OK"]:
                return res
            guid = res["Value"]
            linkPath = "%s/%s" % (datasetName, guid)
            links.append(linkPath)
        res = self.removeLink(links)
        if not res["OK"]:
            return res
        if len(res["Value"]["Successful"]) == len(links):
            return S_OK()
        totalError = ""
        for link, error in res["Value"]["Failed"].items():
            gLogger.error(
                "LcgFileCatalogClient.__removeFilesFromDataset: Failed to remove link", "%s: %s" % (link, error)
            )
            totalError = "%s %s : %s" % (totalError, link, error)
        return S_ERROR(totalError)

    ####################################################################
    #
    # These are the methods required for the admin interface
    #

    def getUserDirectory(self, usernames):
        """Takes a list of users and determines whether their directories already exist"""
        result = getClientCertInfo()
        if not result["OK"]:
            return result
        vo = result["Value"]["VO"]
        usernameDict = {}
        for username in usernames:
            userDirectory = "/%s/user/%s/%s" % (vo, username[0], username)
            usernameDict[userDirectory] = username
        res = self.exists(list(usernameDict))
        if not res["OK"]:
            return res
        failed = {}
        for directory, reason in res["Value"]["Failed"].items():
            failed[usernameDict[directory]] = reason
        successful = {}
        for directory, exists in res["Value"]["Successful"].items():
            successful[usernameDict[directory]] = exists
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def createUserDirectory(self, usernames):
        """Creates the user directory"""
        result = getClientCertInfo()
        if not result["OK"]:
            return result
        vo = result["Value"]["VO"]
        successful = {}
        failed = {}
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        for username in usernames:
            userDirectory = "/%s/user/%s/%s" % (vo, username[0], username)
            res = self.__makeDirs(userDirectory, 0o755)
            if res["OK"]:
                successful[username] = userDirectory
            else:
                failed[username] = res["Message"]
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def removeUserDirectory(self, username):
        """Remove the user directory and remove the user mapping"""
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        res = self.getUserDirectory(username)
        if not res["OK"]:
            return res
        failed = {}
        for username, error in res["Value"]["Failed"].items():
            failed[username] = error
        directoriesToRemove = {}
        successful = {}
        for username, directory in res["Value"]["Successful"].items():
            if not directory:
                successful[username] = True
            else:
                directoriesToRemove[directory] = username
        res = self.removeDirectory(list(directoriesToRemove))
        if not res["OK"]:
            return res
        for directory, error in res["Value"]["Failed"].items():
            failed[directoriesToRemove[directory]] = error
        for directory in res["Value"]["Successful"]:
            successful[directoriesToRemove[directory]] = True
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def changePathOwner(self, paths):
        """Change the ownership of the directory to the user associated to the supplied DN"""
        successful = {}
        failed = {}
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        for dirPath, dn in paths.items():
            res = getDNUserID(dn)
            if not res["OK"]:
                failed[dirPath] = res["Message"]
            else:
                userID = res["Value"]
                res = self.__changeOwner(dirPath, userID)
                if not res["OK"]:
                    failed[dirPath] = res["Message"]
                else:
                    successful[dirPath] = True
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    @checkCatalogArguments
    def changePathMode(self, paths):
        """Change the ownership of the directory to the user associated to the supplied DN"""
        successful = {}
        failed = {}
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        for dirPath, mode in paths.items():
            res = self.__changeMode(dirPath, mode)
            if not res["OK"]:
                failed[dirPath] = res["Message"]
            else:
                successful[dirPath] = True
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def createUserMapping(self, userDNs):
        """Create a user with the supplied DN and return the userID"""
        successful = {}
        failed = {}
        created = self.__openSession()
        if created < 0:
            return S_ERROR("Error opening LFC session")
        for userDN, uid in userDNs.items():
            if not uid:
                uid = -1
            res = addUserDN(uid, userDN)
            if not res["OK"]:
                failed[userDN] = res["Message"]
            else:
                res = getDNUserID(userDN)
                if not res["OK"]:
                    failed[userDN] = res["Message"]
                else:
                    successful[userDN] = res["Value"]
        if created:
            self.__closeSession()
        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    ####################################################################
    #
    # These are the internal methods used for the admin interface
    #

    def __changeOwner(self, lfn, userID):
        return returnCode(lfc.lfc_chown(self.__fullLfn(lfn), userID, -1))

    def __changeMod(self, lfn, mode):
        return returnCode(lfc.lfc_chmod(self.__fullLfn(lfn), mode))

    def __fullLfn(self, lfn):
        return str(self.prefix + lfn)

    # THIS IS NOT YET WORKING



#################################################################
# Class DirectoryListing
# Author: A.T.
# Added 02.03.2015
#################################################################

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import stat

from DIRAC import gConfig
from DIRAC.ConfigurationSystem.Client.Helpers import Registry


class DirectoryListing(object):
    def __init__(self):

        self.entries = []

    def addFile(self, name, fileDict, repDict, numericid):
        """Pretty print of the file ls output"""
        perm = fileDict["Mode"]
        date = fileDict["ModificationDate"]
        # nlinks = fileDict.get('NumberOfLinks',0)
        nreplicas = len(repDict)
        size = fileDict["Size"]
        if "Owner" in fileDict:
            uname = fileDict["Owner"]
        elif "OwnerDN" in fileDict:
            result = Registry.getUsernameForDN(fileDict["OwnerDN"])
            if result["OK"]:
                uname = result["Value"]
            else:
                uname = "unknown"
        else:
            uname = "unknown"
        if numericid:
            uname = str(fileDict["UID"])
        if "OwnerGroup" in fileDict:
            gname = fileDict["OwnerGroup"]
        elif "OwnerRole" in fileDict:
            groups = Registry.getGroupsWithVOMSAttribute("/" + fileDict["OwnerRole"])
            if groups:
                if len(groups) > 1:
                    gname = groups[0]
                    default_group = gConfig.getValue("/Registry/DefaultGroup", "unknown")
                    if default_group in groups:
                        gname = default_group
                else:
                    gname = groups[0]
            else:
                gname = "unknown"
        else:
            gname = "unknown"
        if numericid:
            gname = str(fileDict["GID"])

        self.entries.append(("-" + self.__getModeString(perm), nreplicas, uname, gname, size, date, name))

    def addDirectory(self, name, dirDict, numericid):
        """Pretty print of the file ls output"""
        perm = dirDict["Mode"]
        date = dirDict["ModificationDate"]
        nlinks = 0
        size = 0
        if "Owner" in dirDict:
            uname = dirDict["Owner"]
        elif "OwnerDN" in dirDict:
            result = Registry.getUsernameForDN(dirDict["OwnerDN"])
            if result["OK"]:
                uname = result["Value"]
            else:
                uname = "unknown"
        else:
            uname = "unknown"
        if numericid:
            uname = str(dirDict["UID"])
        if "OwnerGroup" in dirDict:
            gname = dirDict["OwnerGroup"]
        elif "OwnerRole" in dirDict:
            groups = Registry.getGroupsWithVOMSAttribute("/" + dirDict["OwnerRole"])
            if groups:
                if len(groups) > 1:
                    gname = groups[0]
                    default_group = gConfig.getValue("/Registry/DefaultGroup", "unknown")
                    if default_group in groups:
                        gname = default_group
                else:
                    gname = groups[0]
            else:
                gname = "unknown"
        if numericid:
            gname = str(dirDict["GID"])

        self.entries.append(("d" + self.__getModeString(perm), nlinks, uname, gname, size, date, name))

    def addDataset(self, name, datasetDict, numericid):
        """Pretty print of the dataset ls output"""
        perm = datasetDict["Mode"]
        date = datasetDict["ModificationDate"]
        size = datasetDict["TotalSize"]
        if "Owner" in datasetDict:
            uname = datasetDict["Owner"]
        elif "OwnerDN" in datasetDict:
            result = Registry.getUsernameForDN(datasetDict["OwnerDN"])
            if result["OK"]:
                uname = result["Value"]
            else:
                uname = "unknown"
        else:
            uname = "unknown"
        if numericid:
            uname = str(datasetDict["UID"])

        gname = "unknown"
        if "OwnerGroup" in datasetDict:
            gname = datasetDict["OwnerGroup"]
        if numericid:
            gname = str(datasetDict["GID"])

        numberOfFiles = datasetDict["NumberOfFiles"]

        self.entries.append(("s" + self.__getModeString(perm), numberOfFiles, uname, gname, size, date, name))

    def __getModeString(self, perm):
        """Get string representation of the file/directory mode"""

        pstring = ""
        if perm & stat.S_IRUSR:
            pstring += "r"
        else:
            pstring += "-"
        if perm & stat.S_IWUSR:
            pstring += "w"
        else:
            pstring += "-"
        if perm & stat.S_IXUSR:
            pstring += "x"
        else:
            pstring += "-"
        if perm & stat.S_IRGRP:
            pstring += "r"
        else:
            pstring += "-"
        if perm & stat.S_IWGRP:
            pstring += "w"
        else:
            pstring += "-"
        if perm & stat.S_IXGRP:
            pstring += "x"
        else:
            pstring += "-"
        if perm & stat.S_IROTH:
            pstring += "r"
        else:
            pstring += "-"
        if perm & stat.S_IWOTH:
            pstring += "w"
        else:
            pstring += "-"
        if perm & stat.S_IXOTH:
            pstring += "x"
        else:
            pstring += "-"

        return pstring

    def humanReadableSize(self, num, suffix="B"):
        """Translate file size in bytes to human readable

        Powers of 2 are used (1Mi = 2^20 = 1048576 bytes).
        """
        num = int(num)
        for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
            if abs(num) < 1024.0:
                return "%3.1f%s%s" % (num, unit, suffix)
            num /= 1024.0
        return "%.1f%s%s" % (num, "Yi", suffix)

    def printListing(self, reverse, timeorder, sizeorder, humanread):
        """ """
        if timeorder:
            if reverse:
                self.entries.sort(key=lambda x: x[5])
            else:
                self.entries.sort(key=lambda x: x[5], reverse=True)
        elif sizeorder:
            if reverse:
                self.entries.sort(key=lambda x: x[4])
            else:
                self.entries.sort(key=lambda x: x[4], reverse=True)
        else:
            if reverse:
                self.entries.sort(key=lambda x: x[6], reverse=True)
            else:
                self.entries.sort(key=lambda x: x[6])

        # Determine the field widths
        wList = [0] * 7
        for d in self.entries:
            for i in range(7):
                if humanread and i == 4:
                    humanreadlen = len(str(self.humanReadableSize(d[4])))
                    if humanreadlen > wList[4]:
                        wList[4] = humanreadlen
                else:
                    if len(str(d[i])) > wList[i]:
                        wList[i] = len(str(d[i]))

        for e in self.entries:
            size = e[4]
            if humanread:
                size = self.humanReadableSize(e[4])
            print(str(e[0]), end=" ")
            print(str(e[1]).rjust(wList[1]), end=" ")
            print(str(e[2]).ljust(wList[2]), end=" ")
            print(str(e[3]).ljust(wList[3]), end=" ")
            print(str(size).rjust(wList[4]), end=" ")
            print(str(e[5]).rjust(wList[5]), end=" ")
            print(str(e[6]))

    def addSimpleFile(self, name):
        """Add single files to be sorted later"""
        self.entries.append(name)

    def printOrdered(self):
        """print the ordered list"""
        self.entries.sort()
        for entry in self.entries:
            print(entry)



""" VirtualMachineDB class is a front-end to the virtual machines DB

  Life cycle of VMs Images in DB
    * New:       Inserted by Director (Name - Status = New ) if not existing when launching a new instance
    * Validated: Declared by VMMonitoring Server when an Instance reports back correctly
    * Error:     Declared by VMMonitoring Server when an Instance reports back wrong requirements

  Life cycle of VMs Instances in DB
    * New:       Inserted by Director before launching a new instance, to check if image is valid
    * Submitted: Inserted by Director (adding UniqueID) when launches a new instance
    * Wait_ssh_context: Declared by Director for submitted instance wich need later contextualization using ssh
                        (VirtualMachineContextualization will check)
    * Contextualizing:     on the waith_ssh_context path is the next status before Running
    * Running:   Declared by VMMonitoring Server when an Instance reports back correctly
                   (add LastUpdate, publicIP and privateIP)
    * Stopping:  Declared by VMManager Server when an Instance has been deleted outside of the VM
                   (f.e "Delete" button on Browse Instances)
    * Halted:    Declared by VMMonitoring Server when an Instance reports halting
    * Stalled:   Declared by VMManager Server when detects Instance no more running
    * Error:     Declared by VMMonitoring Server when an Instance reports back wrong requirements
                   or reports as running when Halted

  New Instances can be launched by Director if VMImage is not in Error Status.

  Instance UniqueID: for KVM it could be the MAC, for Amazon the returned InstanceID(i-5dec3236),
  or Occi returned the VMID

"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import six

# DIRAC
from DIRAC import gConfig, S_ERROR, S_OK
from DIRAC.Core.Base.DB import DB
from DIRAC.Core.Utilities import Time

__RCSID__ = "$Id$"


class VirtualMachineDB(DB):

    # When checking the Status on the DB it must be one of these values, if not, the last one (Error) is set
    # When declaring a new Status, it will be set to Error if not in the list
    validImageStates = ["New", "Validated", "Error"]
    validInstanceStates = ["New", "Submitted", "Running", "Stopping", "Halted", "Stalled", "Error"]

    # In seconds !
    stallingInterval = 60 * 40

    # When attempting a transition it will be checked if the current state is allowed
    allowedTransitions = {
        "Image": {
            "Validated": ["New", "Validated"],
        },
        "Instance": {
            "Submitted": ["New"],
            "Running": ["Submitted", "Running", "Stalled", "New"],
            "Stopping": ["Running", "Stalled"],
            "Halted": ["New", "Running", "Stopping", "Stalled", "Halted"],
            "Stalled": ["New", "Submitted", "Running"],
        },
    }

    tablesDesc = {}

    tablesDesc["vm_Images"] = {
        "Fields": {
            "VMImageID": "BIGINT UNSIGNED AUTO_INCREMENT NOT NULL",
            "Name": "VARCHAR(255) NOT NULL",
            "Status": "VARCHAR(16) NOT NULL",
            "LastUpdate": "DATETIME",
            "ErrorMessage": 'VARCHAR(255) NOT NULL DEFAULT ""',
        },
        "PrimaryKey": "VMImageID",
    }

    tablesDesc["vm_Instances"] = {
        "Fields": {
            "InstanceID": "BIGINT UNSIGNED AUTO_INCREMENT NOT NULL",
            "RunningPod": "VARCHAR(255) NOT NULL",
            "Name": "VARCHAR(255) NOT NULL",
            "Endpoint": "VARCHAR(255) NOT NULL",
            "UniqueID": 'VARCHAR(255) NOT NULL DEFAULT ""',
            "VMImageID": "INTEGER UNSIGNED NOT NULL",
            "Status": "VARCHAR(32) NOT NULL",
            "LastUpdate": "DATETIME",
            "PublicIP": 'VARCHAR(32) NOT NULL DEFAULT ""',
            "PrivateIP": 'VARCHAR(32) NOT NULL DEFAULT ""',
            "ErrorMessage": 'VARCHAR(255) NOT NULL DEFAULT ""',
            "MaxAllowedPrice": "FLOAT DEFAULT NULL",
            "Uptime": "INTEGER UNSIGNED DEFAULT 0",
            "Load": "FLOAT DEFAULT 0",
            "Jobs": "INTEGER UNSIGNED NOT NULL DEFAULT 0",
        },
        "PrimaryKey": "InstanceID",
        "Indexes": {"Status": ["Status"]},
    }

    tablesDesc["vm_History"] = {
        "Fields": {
            "InstanceID": "INTEGER UNSIGNED NOT NULL",
            "Status": "VARCHAR(32) NOT NULL",
            "Load": "FLOAT NOT NULL",
            "Jobs": "INTEGER UNSIGNED NOT NULL DEFAULT 0",
            "TransferredFiles": "INTEGER UNSIGNED NOT NULL DEFAULT 0",
            "TransferredBytes": "BIGINT UNSIGNED NOT NULL DEFAULT 0",
            "Update": "DATETIME",
        },
        "Indexes": {"InstanceID": ["InstanceID"]},
    }

    #######################
    # VirtualDB constructor
    #######################

    def __init__(self, maxQueueSize=10):
        super(VirtualMachineDB, self).__init__("VirtualMachineDB", "WorkloadManagement/VirtualMachineDB")
        result = self.__initializeDB()
        if not result["OK"]:
            raise Exception("Can't create tables: %s" % result["Message"])

    #######################
    # Public Functions
    #######################

    def checkImageStatus(self, imageName):
        """
        Check Status of a given image
        Will insert a new Image in the DB if it does not exits
        returns:
        S_OK(Status) if Status is valid and not Error
        S_ERROR(ErrorMessage) otherwise
        """
        ret = self.__getImageID(imageName)
        if not ret["OK"]:
            return ret
        return self.__getStatus("Image", ret["Value"])

    def insertInstance(self, uniqueID, imageName, instanceName, endpoint, runningPodName):
        """
        Check Status of a given image
        Will insert a new Instance in the DB
        returns:
        S_OK( InstanceID ) if new Instance is properly inserted
        S_ERROR(ErrorMessage) otherwise
        """
        imageStatus = self.checkImageStatus(imageName)
        if not imageStatus["OK"]:
            return imageStatus

        return self.__insertInstance(uniqueID, imageName, instanceName, endpoint, runningPodName)

    def setInstanceUniqueID(self, instanceID, uniqueID):
        """
        Assign a uniqueID to an instance
        """
        result = self.getInstanceID(uniqueID)
        if result["OK"]:
            return S_ERROR("UniqueID is not unique: %s" % uniqueID)

        result = self._escapeString(uniqueID)
        if not result["OK"]:
            return result
        uniqueID = result["Value"]

        try:
            instanceID = int(instanceID)
        except ValueError:
            return S_ERROR("instanceID has to be a number")

        tableName, _validStates, idName = self.__getTypeTuple("Instance")

        sqlUpdate = "UPDATE `%s` SET UniqueID = %s WHERE %s = %d" % (tableName, uniqueID, idName, instanceID)
        return self._update(sqlUpdate)

    def getInstanceParameter(self, pName, instanceID):
        """Get the instance parameter pName for the given instanceID

        :param str pName: parameter name
        :param str instanceID: instance unique identifier
        :return: S_OK/S_ERROR, parameter value
        """

        tableName, _validStates, idName = self.__getTypeTuple("Instance")

        if pName not in VirtualMachineDB.tablesDesc["vm_Instances"]["Fields"]:
            return S_ERROR("Invalid Instance parameter %s" % pName)

        sqlQuery = "SELECT `%s` FROM `%s` WHERE %s = %s" % (pName, tableName, idName, instanceID)
        result = self._query(sqlQuery)

        if not result["OK"]:
            return result
        value = result["Value"][0][0]

        return S_OK(value)

    def getUniqueID(self, instanceID):
        """
        For a given dirac instanceID get the corresponding cloud endpoint uniqueID
        """
        return self.getInstanceParameter("UniqueID", instanceID)

    def getUniqueIDByName(self, instanceName):
        """Get the cloud provider unique ID corresponding to the DIRAC unique name

        :param str instanceName: VM name
        :return: S_OK/S_ERROR, cloud unique ID as value
        """
        tableName, _validStates, idName = self.__getTypeTuple("Instance")

        sqlQuery = "SELECT UniqueID FROM `%s` WHERE Name = '%s'" % (tableName, instanceName)
        result = self._query(sqlQuery)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Instance not found")
        uniqueID = result["Value"][0][0]

        return S_OK(uniqueID)

    def getInstanceID(self, uniqueID):
        """
        For a given uniqueID of an instance return associated internal InstanceID
        """
        tableName, _validStates, idName = self.__getTypeTuple("Instance")

        result = self.getFields(tableName, [idName], {"UniqueID": uniqueID})
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Unknown %s = %s" % ("UniqueID", uniqueID))
        return S_OK(result["Value"][0][0])

    def declareInstanceSubmitted(self, uniqueID):
        """
        After submission of the instance the Director should declare the submitted Status
        """
        instanceID = self.getInstanceID(uniqueID)
        if not instanceID["OK"]:
            return instanceID
        instanceID = instanceID["Value"]

        status = self.__setState("Instance", instanceID, "Submitted")
        if status["OK"]:
            self.__addInstanceHistory(instanceID, "Submitted")

        return status

    def declareInstanceRunning(self, uniqueID, publicIP, privateIP=""):
        """
        Declares an instance Running and sets its associated info (uniqueID, publicIP, privateIP)
        Returns S_ERROR if:
        - instanceName does not have a "Submitted" or "Contextualizing" entry
        - uniqueID is not unique
        """
        instanceID = self.getInstanceID(uniqueID)
        if not instanceID["OK"]:
            return instanceID
        instanceID = instanceID["Value"]

        # No IPv6 prefix
        publicIP = publicIP.replace("::ffff:", "")

        self.__setInstanceIPs(instanceID, publicIP, privateIP)

        status = self.__setState("Instance", instanceID, "Running")
        if status["OK"]:
            self.__addInstanceHistory(instanceID, "Running")

        return self.getAllInfoForUniqueID(uniqueID)

    def declareInstanceStopping(self, instanceID):
        """
        Mark a VM instance to be stopped.

        Next time the instance's VirtualMachineMonitor checks in for an update
        it will be told to halt.

        :return: S_OK if instance updated, S_ERROR otherwise.
        """
        status = self.__setState("Instance", instanceID, "Stopping")
        if status["OK"]:
            self.__addInstanceHistory(instanceID, "Stopping")
        return status

    def getInstanceStatus(self, instanceID):
        """
        By dirac instanceID
        """
        tableName, validStates, idName = self.__getTypeTuple("Instance")
        if not tableName:
            return S_ERROR("Unknown DB object Instance")

        ret = self.__getStatus("Instance", instanceID)
        if not ret["OK"]:
            return ret

        if not ret["Value"]:
            return S_ERROR("Unknown InstanceID = %s" % (instanceID))

        status = ret["Value"]
        if status not in validStates:
            return self.__setError("Instances", instanceID, "Invalid Status: %s" % status)

        return S_OK(status)

    def recordDBHalt(self, instanceID, load):
        """
        Insert the heart beat info from a halting instance
        Declares "Halted" the instance and the image
        It returns S_ERROR if the status is not OK
        """
        status = self.__setState("Instance", instanceID, "Halted")
        if status["OK"]:
            self.__addInstanceHistory(instanceID, "Halted", load)

        return status

    def declareInstanceHalting(self, uniqueID, load):
        """
        Insert the heart beat info from a halting instance
        Declares "Halted" the instance and the image
        It returns S_ERROR if the status is not OK
        """
        instanceID = self.getInstanceID(uniqueID)
        if not instanceID["OK"]:
            return instanceID
        instanceID = instanceID["Value"]

        status = self.__setState("Instance", instanceID, "Halted")
        if status["OK"]:
            self.__addInstanceHistory(instanceID, "Halted", load)

        return status

    def declareStalledInstances(self):
        """
        Check last Heart Beat for all Running instances and declare them Stalled if older than interval
        """
        oldInstances = self.__getOldInstanceIDs(self.stallingInterval, self.allowedTransitions["Instance"]["Stalled"])
        if not oldInstances["OK"]:
            return oldInstances

        stallingInstances = []

        if not oldInstances["Value"]:
            return S_OK(stallingInstances)

        for instanceID in oldInstances["Value"]:
            instanceID = instanceID[0]
            stalled = self.__setState("Instance", instanceID, "Stalled")
            if not stalled["OK"]:
                continue

            self.__addInstanceHistory(instanceID, "Stalled")
            stallingInstances.append(instanceID)

        return S_OK(stallingInstances)

    def instanceIDHeartBeat(self, uniqueID, load, jobs, transferredFiles, transferredBytes, uptime):
        """
        Insert the heart beat info from a running instance
        It checks the status of the instance and the corresponding image
        Declares "Running" the instance and the image
        It returns S_ERROR if the status is not OK
        """
        instanceID = self.getInstanceID(uniqueID)
        if not instanceID["OK"]:
            return instanceID
        instanceID = instanceID["Value"]

        result = self.__runningInstance(instanceID, load, jobs, transferredFiles, transferredBytes)
        if not result["OK"]:
            return result

        self.__setLastLoadJobsAndUptime(instanceID, load, jobs, uptime)

        status = self.__getStatus("Instance", instanceID)
        if not status["OK"]:
            return result
        status = status["Value"]

        if status == "Stopping":
            return S_OK("stop")
        return S_OK()

    def getEndpointFromInstance(self, uniqueId):
        """
        For a given instance uniqueId it returns the asociated Endpoint in the instance
        table, thus the ImageName of such instance
        """
        tableName, _validStates, _idName = self.__getTypeTuple("Instance")

        endpoint = self.getFields(tableName, ["Endpoint"], {"UniqueID": uniqueId})
        if not endpoint["OK"]:
            return endpoint
        endpoint = endpoint["Value"]

        if not endpoint:
            return S_ERROR("Unknown %s = %s" % ("UniqueID", uniqueId))

        return S_OK(endpoint[0][0])

    def getInstancesByStatus(self, status):
        """
        Get dictionary of Image Names with InstanceIDs in given status
        """
        if status not in self.validInstanceStates:
            return S_ERROR("Status %s is not known" % status)

        # InstanceTuple
        tableName, _validStates, _idName = self.__getTypeTuple("Instance")

        runningInstances = self.getFields(tableName, ["VMImageID", "UniqueID"], {"Status": status})
        if not runningInstances["OK"]:
            return runningInstances
        runningInstances = runningInstances["Value"]

        instancesDict = {}
        imagesDict = {}

        # ImageTuple
        tableName, _validStates, idName = self.__getTypeTuple("Image")

        for imageID, uniqueID in runningInstances:

            if imageID not in imagesDict:

                imageName = self.getFields(tableName, ["Name"], {idName: imageID})
                if not imageName["OK"]:
                    continue
                imagesDict[imageID] = imageName["Value"][0][0]

            if not imagesDict[imageID] in instancesDict:
                instancesDict[imagesDict[imageID]] = []
            instancesDict[imagesDict[imageID]].append(uniqueID)

        return S_OK(instancesDict)

    def getAllInfoForUniqueID(self, uniqueID):
        """
        Get all fields for a uniqueID
        """
        instanceID = self.getInstanceID(uniqueID)
        if not instanceID["OK"]:
            return instanceID
        instanceID = instanceID["Value"]

        instData = self.__getInfo("Instance", instanceID)
        if not instData["OK"]:
            return instData
        instData = instData["Value"]

        imgData = self.__getInfo("Image", instData["VMImageID"])
        if not imgData["OK"]:
            return imgData
        imgData = imgData["Value"]

        return S_OK({"Image": imgData, "Instance": instData})

    #############################
    # Monitoring Public Functions
    #############################

    def getInstancesContent(self, selDict, sortList, start=0, limit=0):
        """
        Function to get the contents of the db
          parameters are a filter to the db
        """
        # Main fields
        tables = ("`vm_Images` AS img", "`vm_Instances` AS inst")
        imageFields = ("VMImageID", "Name")
        instanceFields = (
            "RunningPod",
            "InstanceID",
            "Endpoint",
            "Name",
            "UniqueID",
            "VMImageID",
            "Status",
            "PublicIP",
            "Status",
            "ErrorMessage",
            "LastUpdate",
            "Load",
            "Uptime",
            "Jobs",
        )

        fields = ["img.%s" % f for f in imageFields] + ["inst.%s" % f for f in instanceFields]
        sqlQuery = "SELECT %s FROM %s" % (", ".join(fields), ", ".join(tables))
        sqlCond = ["img.VMImageID = inst.VMImageID"]
        for field in selDict:
            if field in instanceFields:
                sqlField = "inst.%s" % field
            elif field in imageFields:
                sqlField = "img.%s" % field
            elif field in fields:
                sqlField = field
            else:
                continue
            value = selDict[field]
            if isinstance(value, six.string_types):
                value = [str(value)]
            sqlCond.append(
                " OR ".join(["%s=%s" % (sqlField, self._escapeString(str(value))["Value"]) for value in selDict[field]])
            )
        sqlQuery += " WHERE %s" % " AND ".join(sqlCond)
        if sortList:
            sqlSortList = []
            for sorting in sortList:
                if sorting[0] in instanceFields:
                    sqlField = "inst.%s" % sorting[0]
                elif sorting[0] in imageFields:
                    sqlField = "img.%s" % sorting[0]
                elif sorting[0] in fields:
                    sqlField = sorting[0]
                else:
                    continue
                direction = sorting[1].upper()
                if direction not in ("ASC", "DESC"):
                    continue
                sqlSortList.append("%s %s" % (sqlField, direction))
            if sqlSortList:
                sqlQuery += " ORDER BY %s" % ", ".join(sqlSortList)
        if limit:
            sqlQuery += " LIMIT %d,%d" % (start, limit)
        retVal = self._query(sqlQuery)
        if not retVal["OK"]:
            return retVal
        data = []
        # Total records
        for record in retVal["Value"]:
            record = list(record)
            data.append(record)
        totalRecords = len(data)
        sqlQuery = "SELECT COUNT( InstanceID ) FROM %s WHERE %s" % (", ".join(tables), " AND ".join(sqlCond))
        retVal = self._query(sqlQuery)
        if retVal["OK"]:
            totalRecords = retVal["Value"][0][0]
        # return
        return S_OK({"ParameterNames": fields, "Records": data, "TotalRecords": totalRecords})

    def getHistoryForInstanceID(self, instanceId):
        try:
            instanceId = int(instanceId)
        except ValueError:
            return S_ERROR("Instance Id has to be a number!")

        fields = ("Status", "Load", "Update", "Jobs", "TransferredFiles", "TransferredBytes")
        sqlFields = ["`%s`" % f for f in fields]

        sqlQuery = "SELECT %s FROM `vm_History` WHERE InstanceId=%d" % (", ".join(sqlFields), instanceId)
        retVal = self._query(sqlQuery)
        if not retVal["OK"]:
            return retVal
        return S_OK({"ParameterNames": fields, "Records": retVal["Value"]})

    def getInstanceCounters(self, groupField="Status", selDict=None):
        if not selDict:
            selDict = {}
        validFields = VirtualMachineDB.tablesDesc["vm_Instances"]["Fields"]
        if groupField not in validFields:
            return S_ERROR("%s is not a valid field" % groupField)
        sqlCond = []
        for field in selDict:
            if field not in validFields:
                return S_ERROR("%s is not a valid field" % field)
            value = selDict[field]
            if not isinstance(value, (dict, tuple)):
                value = (value,)
            value = [self._escapeString(str(v))["Value"] for v in value]
            sqlCond.append("`%s` in (%s)" % (field, ", ".join(value)))
        sqlQuery = "SELECT `%s`, COUNT( `%s` ) FROM `vm_Instances`" % (groupField, groupField)

        if sqlCond:
            sqlQuery += " WHERE %s" % " AND ".join(sqlCond)
        sqlQuery += " GROUP BY `%s`" % groupField

        result = self._query(sqlQuery)
        if not result["OK"]:
            return result
        return S_OK(dict(result["Value"]))

    def getHistoryValues(self, averageBucket, selDict=None, fields2Get=False, timespan=0):
        if not selDict:
            selDict = {}
        try:
            timespan = max(0, int(timespan))
        except ValueError:
            return S_ERROR("Timespan has to be an integer")

        cumulativeFields = ["Jobs", "TransferredFiles", "TransferredBytes"]
        validDataFields = ["Load", "Jobs", "TransferredFiles", "TransferredBytes"]
        allValidFields = VirtualMachineDB.tablesDesc["vm_History"]["Fields"]

        if not fields2Get:
            fields2Get = list(validDataFields)
        for field in fields2Get:
            if field not in validDataFields:
                return S_ERROR("%s is not a valid data field" % field)

        # paramFields = fields2Get
        try:
            bucketSize = int(averageBucket)
        except ValueError:
            return S_ERROR("Average bucket has to be an integer")

        sqlGroup = "FROM_UNIXTIME(UNIX_TIMESTAMP( `Update` ) - UNIX_TIMESTAMP( `Update` ) mod %d)" % bucketSize
        sqlFields = ["`InstanceID`", sqlGroup]  # + [ "SUM(`%s`)/COUNT(`%s`)" % ( f, f ) for f in fields2Get ]
        for field in fields2Get:
            if field in cumulativeFields:
                sqlFields.append("MAX(`%s`)" % field)
            else:
                sqlFields.append("SUM(`%s`)/COUNT(`%s`)" % (field, field))

        sqlGroup = "%s, InstanceID" % sqlGroup
        paramFields = ["Update"] + fields2Get
        sqlCond = []

        for field in selDict:
            if field not in allValidFields:
                return S_ERROR("%s is not a valid field" % field)
            value = selDict[field]
            if not isinstance(value, (list, tuple)):
                value = (value,)
            value = [self._escapeString(str(v))["Value"] for v in value]
            sqlCond.append("`%s` in (%s)" % (field, ", ".join(value)))
        if timespan > 0:
            sqlCond.append("TIMESTAMPDIFF( SECOND, `Update`, UTC_TIMESTAMP() ) < %d" % timespan)
        sqlQuery = "SELECT %s FROM `vm_History`" % ", ".join(sqlFields)
        if sqlCond:
            sqlQuery += " WHERE %s" % " AND ".join(sqlCond)
        sqlQuery += " GROUP BY %s ORDER BY `Update` ASC" % sqlGroup
        result = self._query(sqlQuery)
        if not result["OK"]:
            return result
        dbData = result["Value"]
        # Need ext?
        requireExtension = set()
        for i in range(len(fields2Get)):
            f = fields2Get[i]
            if f in cumulativeFields:
                requireExtension.add(i)

        if requireExtension:
            rDates = []
            for row in dbData:
                if row[1] not in rDates:
                    rDates.append(row[1])
            vmData = {}
            for row in dbData:
                vmID = row[0]
                if vmID not in vmData:
                    vmData[vmID] = {}
                vmData[vmID][row[1]] = row[2:]
            rDates.sort()

            dbData = []
            for vmID in vmData:
                prevValues = []
                for rDate in rDates:
                    if rDate not in vmData[vmID]:
                        if prevValues:
                            instValues = [rDate]
                            instValues.extend(prevValues)
                            dbData.append(instValues)
                    else:
                        row = vmData[vmID][rDate]
                        prevValues = []
                        for i in range(len(row)):
                            if i in requireExtension:
                                prevValues.append(row[i])
                            else:
                                prevValues.append(0)

                        instValues = [rDate]
                        for i in range(len(row)):
                            instValues.extend(row)
                        dbData.append(instValues)
        else:
            # If we don't require extension just strip vmName
            dbData = [row[1:] for row in dbData]

        # Final sum
        sumData = {}
        for record in dbData:
            recDate = record[0]
            rawData = record[1:]
            if recDate not in sumData:
                sumData[recDate] = [0.0 for f in rawData]
            for i in range(len(rawData)):
                sumData[recDate][i] += float(rawData[i])
        finalData = []
        if len(sumData) > 0:
            firstValues = sumData[sorted(sumData)[0]]
            for date in sorted(sumData):
                finalData.append([date])
                values = sumData[date]
                for i in range(len(values)):
                    if i in requireExtension:
                        finalData[-1].append(max(0, values[i] - firstValues[i]))
                    else:
                        finalData[-1].append(values[i])

        return S_OK({"ParameterNames": paramFields, "Records": finalData})

    def getRunningInstancesHistory(self, timespan=0, bucketSize=900):

        try:
            bucketSize = max(300, int(bucketSize))
        except ValueError:
            return S_ERROR("Bucket has to be an integer")

        try:
            timespan = max(0, int(timespan))
        except ValueError:
            return S_ERROR("Timespan has to be an integer")

        groupby = "FROM_UNIXTIME(UNIX_TIMESTAMP( `Update` ) - UNIX_TIMESTAMP( `Update` ) mod %d )" % bucketSize
        sqlFields = [groupby, "COUNT( DISTINCT( `InstanceID` ) )"]
        sqlQuery = "SELECT %s FROM `vm_History`" % ", ".join(sqlFields)
        sqlCond = ["`Status` = 'Running'"]

        if timespan > 0:
            sqlCond.append("TIMESTAMPDIFF( SECOND, `Update`, UTC_TIMESTAMP() ) < %d" % timespan)
        sqlQuery += " WHERE %s" % " AND ".join(sqlCond)
        sqlQuery += " GROUP BY %s ORDER BY `Update` ASC" % groupby

        return self._query(sqlQuery)

    def getRunningInstancesBEPHistory(self, timespan=0, bucketSize=900):
        try:
            bucketSize = max(300, int(bucketSize))
        except ValueError:
            return S_ERROR("Bucket has to be an integer")
        try:
            timespan = max(0, int(timespan))
        except ValueError:
            return S_ERROR("Timespan has to be an integer")

        groupby = "FROM_UNIXTIME(UNIX_TIMESTAMP( h.`Update` ) - UNIX_TIMESTAMP( h.`Update` ) mod %d )" % bucketSize
        sqlFields = [groupby, " i.Endpoint, COUNT( DISTINCT( h.`InstanceID` ) ) "]
        sqlQuery = "SELECT %s FROM `vm_History` h, `vm_Instances` i" % ", ".join(sqlFields)
        sqlCond = [" h.InstanceID = i.InstanceID AND h.`Status` = 'Running'"]

        if timespan > 0:
            sqlCond.append("TIMESTAMPDIFF( SECOND, `Update`, UTC_TIMESTAMP() ) < %d" % timespan)
        sqlQuery += " WHERE %s" % " AND ".join(sqlCond)
        sqlQuery += " GROUP BY %s , EndPoint ORDER BY `Update` ASC" % groupby

        return self._query(sqlQuery)

    def getRunningInstancesByRunningPodHistory(self, timespan=0, bucketSize=900):
        try:
            bucketSize = max(300, int(bucketSize))
        except ValueError:
            return S_ERROR("Bucket has to be an integer")
        try:
            timespan = max(0, int(timespan))
        except ValueError:
            return S_ERROR("Timespan has to be an integer")

        groupby = "FROM_UNIXTIME(UNIX_TIMESTAMP( h.`Update` ) - UNIX_TIMESTAMP( h.`Update` ) mod %d )" % bucketSize
        sqlFields = [groupby, " i.RunningPod, COUNT( DISTINCT( h.`InstanceID` ) ) "]
        sqlQuery = "SELECT %s FROM `vm_History` h, `vm_Instances` i" % ", ".join(sqlFields)
        sqlCond = [" h.InstanceID = i.InstanceID AND h.`Status` = 'Running'"]

        if timespan > 0:
            sqlCond.append("TIMESTAMPDIFF( SECOND, `Update`, UTC_TIMESTAMP() ) < %d" % timespan)
        sqlQuery += " WHERE %s" % " AND ".join(sqlCond)
        sqlQuery += " GROUP BY %s , RunningPod ORDER BY `Update` ASC" % groupby

        return self._query(sqlQuery)

    def getRunningInstancesByImageHistory(self, timespan=0, bucketSize=900):
        try:
            bucketSize = max(300, int(bucketSize))
        except ValueError:
            return S_ERROR("Bucket has to be an integer")
        try:
            timespan = max(0, int(timespan))
        except ValueError:
            return S_ERROR("Timespan has to be an integer")

        groupby = "FROM_UNIXTIME(UNIX_TIMESTAMP( h.`Update` ) - UNIX_TIMESTAMP( h.`Update` ) mod %d )" % bucketSize
        sqlFields = [groupby, " ins.Name, COUNT( DISTINCT( h.`InstanceID` ) ) "]
        sqlQuery = "SELECT %s FROM `vm_History` h, `vm_Images` img, `vm_Instances` ins" % ", ".join(sqlFields)
        sqlCond = [" h.InstanceID = ins.InstanceID AND img.VMImageID = ins.VMImageID AND h.`Status` = 'Running'"]

        if timespan > 0:
            sqlCond.append("TIMESTAMPDIFF( SECOND, `Update`, UTC_TIMESTAMP() ) < %d" % timespan)
        sqlQuery += " WHERE %s" % " AND ".join(sqlCond)
        sqlQuery += " GROUP BY %s , ins.Name ORDER BY `Update` ASC" % groupby

        return self._query(sqlQuery)

    #######################
    # Private Functions
    #######################

    def __initializeDB(self):
        """
        Create the tables
        """
        tables = self._query("show tables")
        if not tables["OK"]:
            return tables

        tablesInDB = [table[0] for table in tables["Value"]]

        tablesToCreate = {}
        for tableName in self.tablesDesc:
            if tableName not in tablesInDB:
                tablesToCreate[tableName] = self.tablesDesc[tableName]

        return self._createTables(tablesToCreate)

    def __getTypeTuple(self, element):
        """
        return tuple of (tableName, validStates, idName) for object
        """
        # defaults
        tableName, validStates, idName = "", [], ""

        if element == "Image":
            tableName = "vm_Images"
            validStates = self.validImageStates
            idName = "VMImageID"
        elif element == "Instance":
            tableName = "vm_Instances"
            validStates = self.validInstanceStates
            idName = "InstanceID"

        return (tableName, validStates, idName)

    def __insertInstance(self, uniqueID, imageName, instanceName, endpoint, runningPodName):
        """
        Attempts to insert a new Instance for the given Image in a given Endpoint of a runningPodName
        """
        image = self.__getImageID(imageName)
        if not image["OK"]:
            return image
        imageID = image["Value"]

        tableName, validStates, _idName = self.__getTypeTuple("Instance")

        if uniqueID:
            status = "Submitted"
        else:
            status = validStates[0]

        fields = ["UniqueID", "RunningPod", "Name", "Endpoint", "VMImageID", "Status", "LastUpdate"]
        values = [uniqueID, runningPodName, instanceName, endpoint, imageID, status, Time.toString()]

        instance = self.insertFields(tableName, fields, values)
        if not instance["OK"]:
            return instance

        if "lastRowId" in instance:
            self.__addInstanceHistory(instance["lastRowId"], status)
            return S_OK(instance["lastRowId"])

        return S_ERROR("Failed to insert new Instance")

    def __runningInstance(self, instanceID, load, jobs, transferredFiles, transferredBytes):
        """
        Checks image status, set it to running and set instance status to running
        """
        # Check the Image is OK
        imageID = self.__getImageForRunningInstance(instanceID)
        if not imageID["OK"]:
            self.__setError("Instance", instanceID, imageID["Message"])
            return imageID
        imageID = imageID["Value"]

        # Update Instance to Running
        stateInstance = self.__setState("Instance", instanceID, "Running")
        if not stateInstance["OK"]:
            return stateInstance

        # Update Image to Validated
        stateImage = self.__setState("Image", imageID, "Validated")
        if not stateImage["OK"]:
            self.__setError("Instance", instanceID, stateImage["Message"])
            return stateImage

        # Add History record
        self.__addInstanceHistory(instanceID, "Running", load, jobs, transferredFiles, transferredBytes)
        return S_OK()

    def __getImageForRunningInstance(self, instanceID):
        """
        Looks for imageID for a given instanceID.
        Check image Transition to Running is allowed
        Returns:
          S_OK( imageID )
          S_ERROR( Reason )
        """
        info = self.__getInfo("Instance", instanceID)
        if not info["OK"]:
            return info
        info = info["Value"]

        _tableName, _validStates, idName = self.__getTypeTuple("Image")

        imageID = info[idName]

        imageStatus = self.__getStatus("Image", imageID)
        if not imageStatus["OK"]:
            return imageStatus

        return S_OK(imageID)

    def __getOldInstanceIDs(self, secondsIdle, states):
        """
        Return list of instance IDs that have not updated after the given time stamp
        they are required to be in one of the given states
        """
        tableName, _validStates, idName = self.__getTypeTuple("Instance")

        sqlCond = []
        sqlCond.append("TIMESTAMPDIFF( SECOND, `LastUpdate`, UTC_TIMESTAMP() ) > % d" % secondsIdle)
        sqlCond.append('Status IN ( "%s" )' % '", "'.join(states))

        sqlSelect = "SELECT %s from `%s` WHERE %s" % (idName, tableName, " AND ".join(sqlCond))

        return self._query(sqlSelect)

    def __getSubmittedInstanceID(self, imageName):
        """
        Retrieve and InstanceID associated to a submitted Instance for a given Image
        """
        tableName, _validStates, idName = self.__getTypeTuple("Image")

        imageID = self.getFields(tableName, [idName], {"Name": imageName})
        if not imageID["OK"]:
            return imageID
        imageID = imageID["Value"]

        if not imageID:
            return S_ERROR("Unknown Image = %s" % imageName)

        # FIXME: <> is obsolete
        if len(imageID) != 1:
            return S_ERROR('Image name "%s" is not unique' % imageName)

        imageID = imageID[0][0]
        imageIDName = idName

        tableName, _validStates, idName = self.__getTypeTuple("Instance")

        instanceID = self.getFields(tableName, [idName], [imageIDName, "Status"], {imageID: "Submitted"})
        if not instanceID["OK"]:
            return instanceID
        instanceID = instanceID["Value"]

        if not instanceID:
            return S_ERROR('No Submitted instance of "%s" found' % imageName)

        return S_OK(instanceID[0][0])

    def __setState(self, element, iD, state):
        """
        Attempt to set element in state, checking if transition is allowed
        """

        knownStates = self.allowedTransitions[element].keys()
        if state not in knownStates:
            return S_ERROR("Transition to %s not possible" % state)

        allowedStates = self.allowedTransitions[element][state]

        currentState = self.__getStatus(element, iD)
        if not currentState["OK"]:
            return currentState
        currentState = currentState["Value"]

        if currentState not in allowedStates:
            msg = "Transition ( %s -> %s ) not allowed" % (currentState, state)
            if currentState == "Halted":
                val_state = "halt"
            elif currentState == "Stopping":
                val_state = "stop"
            else:
                val_state = currentState
            return {"OK": False, "Message": msg, "State": val_state}

        tableName, _validStates, idName = self.__getTypeTuple(element)

        if currentState == state:
            sqlUpdate = "UPDATE `%s` SET LastUpdate = UTC_TIMESTAMP() WHERE %s = %s" % (tableName, idName, iD)

        else:
            sqlUpdate = 'UPDATE `%s` SET Status = "%s", LastUpdate = UTC_TIMESTAMP() WHERE %s = %s' % (
                tableName,
                state,
                idName,
                iD,
            )

        ret = self._update(sqlUpdate)
        if not ret["OK"]:
            return ret
        return S_OK(state)

    def __setInstanceIPs(self, instanceID, publicIP, privateIP):
        """
        Update parameters for an instanceID reporting as running
        """
        values = self._escapeValues([publicIP, privateIP])
        if not values["OK"]:
            return S_ERROR("Cannot escape values: %s" % str(values))
        publicIP, privateIP = values["Value"]

        tableName, _validStates, idName = self.__getTypeTuple("Instance")
        sqlUpdate = "UPDATE `%s` SET PublicIP = %s, PrivateIP = %s WHERE %s = %s" % (
            tableName,
            publicIP,
            privateIP,
            idName,
            instanceID,
        )

        return self._update(sqlUpdate)

    def __getImageID(self, imageName):
        """
        For a given imageName return corresponding ID
        Will insert the image in New Status if it does not exits,
        """
        tableName, validStates, idName = self.__getTypeTuple("Image")
        imageID = self.getFields(tableName, [idName], {"Name": imageName})
        if not imageID["OK"]:
            return imageID
        imageID = imageID["Value"]

        if len(imageID) > 1:
            return S_ERROR('Image name "%s" is not unique' % imageName)
        if len(imageID) == 0:
            # The image does not exits in DB, has to be inserted
            imageID = 0
        else:
            # The image exits in DB, has to match
            imageID = imageID[0][0]

        if imageID:
            ret = self.getFields(tableName, [idName], {"Name": imageName})
            if not ret["OK"]:
                return ret
            if not ret["Value"]:
                return S_ERROR('Image "%s" in DB but it does not match' % imageName)
            else:
                return S_OK(imageID)

        ret = self.insertFields(
            tableName, ["Name", "Status", "LastUpdate"], [imageName, validStates[0], Time.toString()]
        )

        if ret["OK"] and "lastRowId" in ret:

            rowID = ret["lastRowId"]

            ret = self.getFields(tableName, [idName], {"Name": imageName})
            if not ret["OK"]:
                return ret

            if not ret["Value"] or rowID != ret["Value"][0][0]:
                result = self.__getInfo("Image", rowID)
                if result["OK"]:
                    image = result["Value"]
                    self.log.error('Trying to insert Name: "%s"' % (imageName))
                    self.log.error('But inserted     Name: "%s"' % (image["Name"]))
                return self.__setError("Image", rowID, "Failed to insert new Image")
            return S_OK(rowID)

        return S_ERROR("Failed to insert new Image")

    def __addInstanceHistory(self, instanceID, status, load=0.0, jobs=0, transferredFiles=0, transferredBytes=0):
        """
        Insert a History Record
        """
        try:
            load = float(load)
        except ValueError:
            return S_ERROR("Load has to be a float value")
        try:
            jobs = int(jobs)
        except ValueError:
            return S_ERROR("Jobs has to be an integer value")
        try:
            transferredFiles = int(transferredFiles)
        except ValueError:
            return S_ERROR("Transferred files has to be an integer value")

        self.insertFields(
            "vm_History",
            ["InstanceID", "Status", "Load", "Update", "Jobs", "TransferredFiles", "TransferredBytes"],
            [instanceID, status, load, Time.toString(), jobs, transferredFiles, transferredBytes],
        )
        return

    def __setLastLoadJobsAndUptime(self, instanceID, load, jobs, uptime):
        if not uptime:
            sqlQuery = (
                "SELECT MAX( UNIX_TIMESTAMP( `Update` ) ) - MIN( UNIX_TIMESTAMP( `Update` ) )"
                " FROM `vm_History` WHERE InstanceID = %d GROUP BY InstanceID" % instanceID
            )
            result = self._query(sqlQuery)
            if result["OK"] and len(result["Value"]) > 0:
                uptime = int(result["Value"][0][0])
        sqlUpdate = "UPDATE `vm_Instances` SET `Uptime` = %d, `Jobs`= %d, `Load` = %f WHERE `InstanceID` = %d" % (
            uptime,
            jobs,
            load,
            instanceID,
        )
        self._update(sqlUpdate)
        return S_OK()

    def __getInfo(self, element, iD):
        """
        Return dictionary with info for Images and Instances by ID
        """
        tableName, _validStates, idName = self.__getTypeTuple(element)
        if not tableName:
            return S_ERROR("Unknown DB object: %s" % element)

        fields = self.tablesDesc[tableName]["Fields"]
        ret = self.getFields(tableName, fields, {idName: iD})
        if not ret["OK"]:
            return ret
        if not ret["Value"]:
            return S_ERROR("Unknown %s = %s" % (idName, iD))

        data = {}
        values = ret["Value"][0]
        fields = list(fields.keys())
        return S_OK(dict(zip(fields, values)))

    def __getStatus(self, element, iD):
        """
        Check and return status of Images and Instances by ID
        :return: S_OK(tuple(status(int), message(str))) or S_ERROR(error(str))
        """
        tableName, validStates, idName = self.__getTypeTuple(element)
        if not tableName:
            return S_ERROR("Unknown DB object: %s" % element)

        ret = self.getFields(tableName, ["Status", "ErrorMessage"], {idName: iD})
        if not ret["OK"]:
            return ret

        if not ret["Value"]:
            return S_ERROR("Unknown %s = %s" % (idName, iD))

        status, msg = ret["Value"][0]
        if status not in validStates:
            return self.__setError(element, iD, "Invalid Status: %s" % status)
        if status == validStates[-1]:
            return S_ERROR(msg)

        return S_OK(status)

    def __setError(self, element, iD, reason):
        """ """
        (tableName, validStates, idName) = self.__getTypeTuple(element)
        if not tableName:
            return S_ERROR("Unknown DB object: %s" % element)

        sqlUpdate = 'UPDATE `%s` SET Status = "%s", ErrorMessage = "%s", LastUpdate = UTC_TIMESTAMP() WHERE %s = %s'
        sqlUpdate = sqlUpdate % (tableName, validStates[-1], reason, idName, iD)
        ret = self._update(sqlUpdate)
        if not ret["OK"]:
            return ret

        return S_ERROR(reason)
